

from __future__ import print_function
from time import time
import cProfile
import pstats

import flipper

NT_TYPE_PERIODIC = flipper.kernel.encoding.NT_TYPE_PERIODIC
NT_TYPE_REDUCIBLE = flipper.kernel.encoding.NT_TYPE_REDUCIBLE
NT_TYPE_PSEUDO_ANOSOV = flipper.kernel.encoding.NT_TYPE_PSEUDO_ANOSOV

def main(verbose=False):
	if verbose: print('Running encoding profile.')
	
	start_time = time()
	# Add more tests here.
	# These need to be changed if the standard example triangulations ever change.
	num_isometries = [
		('S_0_4', 2),
		('S_1_1', 6),
		('S_1_2', 4),
		('S_2_1', 2),
		('S_3_1', 2),
		('E_12', 12),
		('E_24', 24),
		('E_36', 36)
		]
	
	# Check that every triangulation has the correct number of isometries to itself.
	for surface, num_isoms in num_isometries:
		if verbose: print('Checking: %s' % surface)
		S = flipper.load(surface)
		T = S.triangulation
		T.self_isometries()
		T2 = flipper.triangulation_from_iso_sig(T.iso_sig())
		T.is_isometric_to(T2)
	
	return time() - start_time

if __name__ == '__main__':
	# print(main(verbose=True))
	#pstats.Stats(cProfile.Profile().run('main()')).strip_dirs().sort_stats('time').print_callers(20)
	#pstats.Stats(cProfile.Profile().run('main()')).strip_dirs().sort_stats('cumtime').print_callees(20)
	pstats.Stats(cProfile.Profile().run('main()')).strip_dirs().sort_stats('time').print_stats(30)





''' A module for representing permutations in Sym(n).

Provides one class: Permutation.

There are also two helper functions: Id_permutation and cyclic_permutation. '''

import flipper

from itertools import permutations, combinations

class Permutation(object):
	''' This represents a permutation in Sym(n). '''
	def __init__(self, permutation):
		assert(isinstance(permutation, (list, tuple)))
		assert(all(isinstance(entry, flipper.IntegerType) for entry in permutation))
		assert(set(permutation) == set(range(len(permutation))))
		
		self.permutation = tuple(permutation)
	def __repr__(self):
		return str(self)
	def __str__(self):
		return self.compressed_string()
	def compressed_string(self):
		''' Return this permutation as a single concatenated string. '''
		
		return ''.join(str(p) for p in self.permutation)
	
	def __iter__(self):
		return iter(self.permutation)
	def __call__(self, index):
		return self.permutation[index]
	def __len__(self):
		return len(self.permutation)
	def __hash__(self):
		return hash(self.permutation)
	def __mul__(self, other):
		assert(isinstance(other, Permutation))
		assert(len(self) == len(other))
		
		return Permutation([self(other(i)) for i in range(len(self))])
	def __eq__(self, other):
		return self.permutation == other.permutation
	def __ne__(self, other):
		return not (self.permutation == other.permutation)
	def inverse(self):
		''' Return the inverse of this permutation. '''
		
		return Permutation([j for i in range(len(self)) for j in range(len(self)) if self(j) == i])
	def is_even(self):
		''' Return if this permutation is even. '''
		
		return len([(i, j) for j, i in combinations(range(len(self)), 2) if self(j) > self(i)]) % 2 == 0
	def order(self):
		''' Return the order of this permutation. '''
		
		id_perm = id_permutation(len(self))
		order = 1
		product = self
		while product != id_perm:
			product = product * self
			order += 1
		return order
	def embed(self, new_n):
		''' Return the inclusion of this permutation into Sym(new_n).
		
		new_n must be at least len(self). '''
		
		assert(new_n >= len(self))  # Cannot embed permutation into smaller symmetric group.
		
		return Permutation(list(self.permutation) + list(range(len(self), new_n)))
	def matrix(self):
		''' Return the corresponding permutation matrix.
		
		That is, a matrix M such that M * e_i == e_{self[i]}. '''
		
		return flipper.kernel.Matrix([[1 if i == self(j) else 0 for j in range(len(self))] for i in range(len(self))])

#### Some special Permutations we know how to build.

def id_permutation(n):
	''' Return the identity permutation in Sym(n). '''
	
	return Permutation(range(n))

def cyclic_permutation(cycle, n):
	''' Return the cyclic permutation sending 1 |--> cycle in Sym(n). '''
	
	return Permutation([(cycle + i) % n for i in range(n)])

def all_permutations(n, odd=True, even=True):
	''' Return a list containing all permutations on n elements.
	
	If even is False then even permutations are omitted. If odd is False
	them odd permutations are omitted. '''
	
	all_perms = [Permutation(perm) for perm in permutations(range(n), n)]
	return [perm for perm in all_perms if (odd and not perm.is_even()) or (even and perm.is_even())]

def permutation_from_pair(a, to_a, b, to_b):
	''' Return the odd permutation in Sym(4) which sends a to to_a and b to to_b. '''
	
	try:
		c, d = set(range(4)) - set([a, b])
		to_c, to_d = set(range(4)) - set([to_a, to_b])
		
		p = {a: to_a, b: to_b, c: to_c, d: to_d}
		P = Permutation([p[i] for i in range(4)])
		if not P.is_even(): return P
		
		p = {a: to_a, b: to_b, c: to_d, d: to_c}
		P = Permutation([p[i] for i in range(4)])
		return P
	except IndexError:
		raise ValueError('Does not represent a gluing.')





try:
	import Tkinter as TK
	import tkFont as TK_FONT
except ImportError:  # Python 3.
	import tkinter as TK
	import tkinter.font as TK_FONT

try:
	import ttk as TTK
except ImportError:  # Python 3.
	from tkinter import ttk as TTK

RENDER_LAMINATION_FULL = 'Full'
RENDER_LAMINATION_W_TRAIN_TRACK = 'Weighted train track'
RENDER_LAMINATION_C_TRAIN_TRACK = 'Compressed train track'
LABEL_EDGES_NONE = 'None'
LABEL_EDGES_INDEX = 'Index'
LABEL_EDGES_GEOMETRIC = 'Geometric'
LABEL_EDGES_ALGEBRAIC = 'Algebraic'
SIZE_SMALL, SIZE_MEDIUM, SIZE_LARGE = 0, 1, 2
# SIZE_XLARGE = 3

class Options(object):
	def __init__(self, parent):
		self.parent = parent
		self.application_font = TK_FONT.Font(family='TkDefaultFont', size=10)
		self.canvas_font = TK_FONT.Font(family='TkDefaultFont', size=10)
		
		self.render_lamination_var = TK.StringVar(value=RENDER_LAMINATION_FULL)
		self.show_internals_var = TK.BooleanVar(value=False)
		self.show_orientations_var = TK.BooleanVar(value=False)
		self.straight_laminations_var = TK.BooleanVar(value=False)
		self.projectivise_var = TK.BooleanVar(value=False)
		self.label_edges_var = TK.StringVar(value=LABEL_EDGES_NONE)
		self.size_var = TK.IntVar(value=SIZE_SMALL)
		
		self.render_lamination = RENDER_LAMINATION_FULL
		self.show_internals = False
		self.show_orientations = False
		self.straight_laminations = False
		self.label_edges = LABEL_EDGES_NONE
		self.projectivise = False
		self.line_size = 2
		self.dot_size = 3
		self.arrow_shape = (12, 15, 5)
		
		# Set it so that self.update() will be called whenever these variables are changed.
		self.render_lamination_var.trace('w', self.update)
		self.show_internals_var.trace('w', self.update)
		self.show_orientations_var.trace('w', self.update)
		self.straight_laminations_var.trace('w', self.update)
		self.projectivise_var.trace('w', self.update)
		self.label_edges_var.trace('w', self.update)
		self.size_var.trace('w', self.update)
		
		# Drawing parameters.
		self.epsilon = 10
		self.float_error = 0.001
		
		self.vertex_buffer = 0.2  # Must be in (0, 0.5)
		self.zoom_fraction = 0.8 # Must be in (0, 1)
	
	def update(self, *args):
		self.render_lamination = str(self.render_lamination_var.get())
		self.show_internals = bool(self.show_internals_var.get())
		self.show_orientations = bool(self.show_orientations_var.get())
		self.straight_laminations = bool(self.straight_laminations_var.get())
		self.projectivise = bool(self.projectivise_var.get())
		self.label_edges = str(self.label_edges_var.get())
		if self.size_var.get() == SIZE_SMALL:
			self.line_size = 2
			self.dot_size = 3
			self.arrow_shape = (12, 15, 5)
			self.application_font.configure(size=10)
			self.canvas_font.configure(size=10)
		elif self.size_var.get() == SIZE_MEDIUM:
			self.line_size = 4
			self.dot_size = 5
			self.arrow_shape = (16, 20, 6)
			self.application_font.configure(size=11)
			self.canvas_font.configure(size=12)
		elif self.size_var.get() == SIZE_LARGE:
			self.line_size = 6
			self.dot_size = 7
			self.arrow_shape = (24, 30, 9)
			self.application_font.configure(size=12)
			self.canvas_font.configure(size=14)
		
		self.parent.treeview_objects.tag_configure('txt', font=self.application_font)
		TTK.Style().configure('Treeview', font=self.application_font)
		self.parent.redraw()





import sys
import argparse
import flipper.application

start = flipper.application.start

if __name__ == '__main__':
	parser = argparse.ArgumentParser(description='Flipper GUI')
	parser.add_argument('load', nargs='?', help='path to load from when starting')
	args = parser.parse_args()
	
	start(load_from=args.load)





''' A module for representing and manipulating intervals.

Provides one class: Interval. '''

import flipper

from math import log10 as log
from math import ceil

# This class represents the interval (lower / 10^precision, upper / 10^precision).

# For an interval I let acc(I) denote the accuracy of I, that is
#	acc(I) := self.precision - int(log(self.upper - self.lower)).
# For an integer x let log+(x) := log(max(abs(x), 1)) and for an interval I
# let log+(I) := max(log+(I.lower), log+(I.upper), I.precision) - I.precision.

# Suppose that x is an integer and that I and J are intervals and that
# m := min(acc(I), acc(J)).
# Then we obtain the following bounds:
#	acc(I + J) >= m - 1,
#	acc(I * J) >= m - log(I.lower + J.lower + 1)
#	acc(I / J) >= m - log+(J)
#	acc(x * I) >= acc(I) - log+(x)

INFTY = float('inf')

class Interval(object):
	''' This represents a closed interval [lower / 10**precision, upper / 10**precision].
	
	>>> w = Interval.from_string('0.10')
	>>> x = Interval.from_string('10000.0')
	>>> y = Interval.from_string('1.14571')
	>>> z = Interval.from_string('1.00000')
	>>> a = Interval.from_string('-1.200000')
	>>> b = Interval.from_string('1.4142135623')
	>>> a
	-1.2001?
	'''
	def __init__(self, lower, upper, precision):
		assert(isinstance(lower, flipper.IntegerType))
		assert(isinstance(upper, flipper.IntegerType))
		assert(isinstance(precision, flipper.IntegerType))
		if lower > upper: raise ValueError('Interval is empty.')
		
		self.lower = lower
		self.upper = upper
		self.precision = precision
		# The width of this interval is at most 10^-self.accuracy.
		# That is, this interval defines a number correct to self.accuracy decimal places.
		self.accuracy = INFTY if self.upper == self.lower else self.precision - int(ceil(log(self.upper - self.lower)))
	
	@classmethod
	def from_string(cls, string):
		''' A short way of constructing Intervals from a string. '''
		
		i, r = string.split('.') if '.' in string else (string, '')
		x = int(i + r)
		return cls(x-1, x+1, len(r))
	
	@classmethod
	def from_int(cls, integer, accuracy=None):
		''' A short way of constructing Intervals from an integer. '''
		
		if accuracy is None:
			return cls(integer, integer, 0)
		else:
			accuracy = int(accuracy) + 1
			return cls(integer*10**accuracy - 1, integer*10**accuracy+1, accuracy)
	
	def __repr__(self):
		return str(self)
	def __str__(self):
		return self.approximate_string(6)
	def __float__(self):
		return float(self.approximate_string(30)[:-1])
	def __int__(self):
		# Return the largest integer less than or equal to all values in self.
		return self.lower // 10**self.precision
	def __hash__(self):
		return hash(self.tuple())
	
	def approximate_string(self, accuracy=None):
		''' Return a string approximating the centre of this interval. '''
		
		if accuracy is None or accuracy > self.accuracy:
			accuracy = self.precision if self.accuracy == INFTY else self.accuracy-1
		I = self.change_denominator(accuracy)
		v = (I.lower + I.upper) // 2
		s = str(v).zfill(accuracy + (2 if v < 0 else 1))
		return '%s.%s?' % (s[:-I.precision], s[-I.precision:])
	def tuple(self):
		''' Return the triple describing this interval. '''
		
		return (self.lower, self.upper, self.precision)
	def log_plus(self):
		''' Return max(log(x), 0) over all x in self. '''
		return max(log(max(abs(self.upper), abs(self.lower), 1)) - self.precision, 0) + 1
	
	def change_denominator(self, new_denominator):
		''' Return a this interval over a new denominator. '''
		
		assert(isinstance(new_denominator, flipper.IntegerType))
		
		d = new_denominator - self.precision
		if d > 0:
			return Interval(self.lower * 10**d, self.upper * 10**d, new_denominator)
		elif d == 0:
			return self
		elif d < 0:
			return Interval((self.lower // 10**(-d)) - 1, (self.upper // 10**(-d)) + 1, new_denominator)
	def change_accuracy(self, new_accuracy):
		''' Return a new interval with the given accuracy.
		
		The new_accuracy must be at most self.accuracy. '''
		
		assert(isinstance(new_accuracy, flipper.IntegerType))
		
		if self.accuracy == INFTY:
			return self
		else:
			assert(new_accuracy <= self.accuracy)
			return self.change_denominator(self.precision - (self.accuracy - new_accuracy))
	def simplify(self):
		''' Return a simpler interval with accuracy at least self.accuracy-1. '''
		
		# As self.change_denominator divides outwards then shifts by +/-1 we need
		# to work with at least two more digits.
		if self.accuracy > 0 and self.precision > self.accuracy + 2:
			I = self.change_denominator(self.accuracy + 2)
		else:
			I = self
		
		assert(I.accuracy >= self.accuracy - 1)
		return I
	
	def __contains__(self, other):
		if isinstance(other, Interval):
			return self.lower <= other.lower and other.upper <= self.upper
		elif isinstance(other, flipper.IntegerType):
			return self.lower <= other * 10**self.precision <= self.upper
		else:
			return NotImplemented
	def __neg__(self):
		return Interval(-self.upper, -self.lower, self.precision)
	def __abs__(self):
		return self if self >= 0 else -self
	def __abs__(self):
		if self.lower > 0:  # Interval is entirely positive.
			return self
		elif self.upper < 0:  # Interval is entirely negative.
			return -self
		if self.lower <= 0 <= self.upper:  # Interval straddles 0.
			return Interval(0, max(-self.lower, self.upper), self.precision)
	def __add__(self, other):
		if isinstance(other, Interval):
			common_precision = max(self.precision, other.precision)
			P, Q = self.change_denominator(common_precision), other.change_denominator(common_precision)
			new_lower = P.lower + Q.lower
			new_upper = P.upper + Q.upper
			return Interval(new_lower, new_upper, common_precision)
		elif isinstance(other, flipper.IntegerType):
			return self + Interval(other, other, 0)
		else:
			return NotImplemented
	def __radd__(self, other):
		return self + other
	def __sub__(self, other):
		if isinstance(other, Interval):
			common_precision = max(self.precision, other.precision)
			P, Q = self.change_denominator(common_precision), other.change_denominator(common_precision)
			new_lower = P.lower - Q.upper
			new_upper = P.upper - Q.lower
			return Interval(new_lower, new_upper, common_precision)
		elif isinstance(other, flipper.IntegerType):
			return self - Interval(other, other, 0)
		else:
			return NotImplemented
	def __rsub__(self, other):
		return -(self - other)
	def __mul__(self, other):
		if isinstance(other, Interval):
			values = [
				self.lower * other.lower,
				self.upper * other.lower,
				self.lower * other.upper,
				self.upper * other.upper
				]
			return Interval(min(values), max(values), self.precision + other.precision)
		elif isinstance(other, flipper.IntegerType):
			values = [self.lower * other, self.upper * other]
			return Interval(min(values), max(values), self.precision)
		else:
			return NotImplemented
	def __rmul__(self, other):
		return self * other
	def __pow__(self, power):
		if power == 0:
			return Interval(1, 1, 0)
		elif power == 1:
			return self
		elif power > 1:
			sqrt = self**(power//2)
			square = sqrt * sqrt
			if power % 2 == 1:
				return self * square
			else:
				return square
	def __div__(self, other):
		return self.__truediv__(other)
	def __truediv__(self, other):
		if isinstance(other, Interval):
			if 0 in other:
				raise ZeroDivisionError  # Denominator contains 0.
			
			# I suspect that there is a lot of optimisation that could be done here.
			# In particular, if we are willing to make the other interval wider we
			# might be able to avoid an expensive division, hopefully without giving
			# up too much accuracy.
			common_precision = max(self.precision, other.precision)
			P, Q = self.change_denominator(common_precision), other.change_denominator(common_precision)
			values = [
				P.lower * 10**common_precision // Q.lower,
				P.upper * 10**common_precision // Q.lower,
				P.lower * 10**common_precision // Q.upper,
				P.upper * 10**common_precision // Q.upper
				]
			# Note: Division always rounds down so we will add +1 to the upper bound to be safe.
			return Interval(min(values), max(values) + 1, common_precision)
		elif isinstance(other, flipper.IntegerType):
			values = [self.lower // other, self.upper // other]
			return Interval(min(values), max(values), self.precision)
		else:
			return NotImplemented
	
	def midpoint(self, magnitude=10):
		''' Return a smaller interval containing the midpoint of this interval. '''
		
		assert(isinstance(magnitude, flipper.IntegerType))
		
		m = (10**magnitude) * (self.lower + self.upper) // 2
		return Interval(m-1, m+1, self.precision+magnitude)
	def intersect(self, other):
		''' Return the intersection of this interval with other. '''
		
		assert(isinstance(other, Interval))
		
		common_precision = max(self.precision, other.precision)
		P, Q = self.change_denominator(common_precision), other.change_denominator(common_precision)
		return Interval(max(P.lower, Q.lower), min(P.upper, Q.upper), common_precision)
	
	def subdivide(self):
		''' Return a list of smaller intervals covering this interval. '''
		
		d = self.upper - self.lower
		if d > 1:
			k = 1 + d // 10
			return [Interval(self.lower + i, min(self.lower + i + k, self.upper), self.precision) for i in range(0, d, k)]
		elif d == 1:
			return [Interval(10*self.lower + i, 10*self.lower + i+1, self.precision + 1) for i in range(10)]
		elif d == 0:
			return [Interval(self.lower, self.upper, self.precision)]
	
	def polynomial(self, degree, scale):
		''' Return a polynomial of at most the specified degree and height with a root (hopefully) close to the center of this interval. '''
		
		# This needs rechecking against the literature to ensure that M.LLL() works correctly.
		M = flipper.kernel.id_matrix(degree+1).join(
			flipper.kernel.Matrix([[int(scale * self**i) for i in range(degree+1)]])).transpose()
		
		# Should this also appear in the symbolic library? Sage is much faster than us at LLL.
		
		return flipper.kernel.Polynomial(M.LLL()[0][:-1])

def doctest_globs():
	''' Return the globals needed to run doctest on this module. '''
	
	w = Interval.from_string('0.10')
	x = Interval.from_string('10000.0')
	y = Interval.from_string('1.14571')
	z = Interval.from_string('1.00000')
	a = Interval.from_string('-1.200000')
	b = Interval.from_string('1.4142135623')
	
	return {'w': w, 'x': x, 'y': y, 'z': z, 'a': a, 'b': b}





from __future__ import print_function

import flipper

def main(verbose=False):
	if verbose: print('Running lamination tests.')
	
	# Add more tests here.
	tests = [
		('S_1_1', 'a'),
		('S_1_2', 'a'),
		('S_1_2', 'b'),
		('S_1_2', 'c'),
		('S_1_2', 'aB'),
		('S_1_2', 'bbaCBAaBabcABB'),
		('S_1_2', 'aCBACBacbaccbAaAAaBBcCcBBcCaBaaaABBabBcaBbCBCbaaa'),
		('S_2_1', 'aaabcd'),
		# ('E_12', 'aaaaBBc'),  # Really slow.
		# ('E_12', 'aaBaaBBc)'  # Really slow.
		# ('E_12', 'aaaaBBaBaBc')  # Really slow useful for profiling. Current best time 102s.
		]
	
	try:
		for surface, word in tests:
			if verbose: print(word)
			S = flipper.load(surface)
			mapping_class = S.mapping_class(word)
			mapping_class.invariant_lamination()
	except flipper.AssumptionError:
		pass  # mapping_class is not pseudo-Anosov.
	
	return True

if __name__ == '__main__':
	print(main(verbose=True))





from __future__ import print_function

import flipper

def main(verbose=False):
	if verbose: print('Running number field tests.')
	
	N = flipper.kernel.NumberField.from_tuple([-2, 0, 1], '1.4142135623')  # QQ(sqrt(2)).
	x = N.lmbda  # sqrt(2)
	
	tests = [
		x * x == 2
		]
	
	return all(tests)

if __name__ == '__main__':
	print(main(verbose=True))





from __future__ import print_function

import flipper
import sys

def main(length):
	S = flipper.load('S_1_1')
	buckets = []  # All the different conjugacy classes that we have found.
	# We could order the buckets by something, say dilatation.
	for index, word in enumerate(S.all_words(length)):
		h = S.mapping_class(word)
		# Currently, we can only determine conjugacy classes for
		# pseudo-Anosovs, so we had better filter by them.
		if h.is_pseudo_anosov():
			# Check if this is conjugate to a mapping class we have seen.
			for bucket in buckets:
				# Conjugacy is transitive, so we only have to bother checking
				# if h is conjugate to the first entry in the bucket.
				if bucket[0].is_conjugate_to(h):
					bucket.append(h)
					break
			else:  # We have found a new conjugacy class.
				buckets.append([h])
		print('\r%d words in %d conjugacy classes.' % (index, len(buckets)), end='')
		sys.stdout.flush()
	
	return buckets

if __name__ == '__main__':
	import argparse
	parser = argparse.ArgumentParser(description='Group mapping classes into conjugacy classes.')
	parser.add_argument('--length', type=int, default=6, help='length of words to generate')
	parser.add_argument('--show', action='store_true', default=False, help='show the source code of this example and exit')
	args = parser.parse_args()
	
	if args.show:
		print(open(__file__, 'r').read())
	else:
		main(args.length)





from __future__ import print_function

import flipper

from time import time

# This example shows that even for atypical mapping classes Encoding.invariant_lamination() is still fast.

def main():
	times = {}
	
	examples = [
		('S_2_1', 'AEeadfaCEeCdEBfbCDFC'),  # Cannot estimate!
		('S_2_1', 'abDeFaabDeFaabDeFaAEeadfaCEeCdEBfbCDFCAfEdBAAfEdBAAfEdBA'),  # Reducing curve [4, 12, 20, 44, 24, 40, 28, 44, 32] - slow!
		('S_2_1', 'aFcE'),  # Dim 2 eigenspace
		('S_2_1', 'AEeadfaCEeCdEBfbCDFC'),  # Word is reducible (reducing curve has weight ~ 6000).
		('S_2_1', 'aDefFecDBdFCcACDcCdF'),  # 12 iterates?
		('S_2_1', 'BcEC'),  # Is reducible but (unlike BC) mixes every edge just a little so growth is super slow.
		('S_2_1', 'ebbFaBDECFbBCFFbFeCa'),  # First iteration has no real roots.
		('S_2_1', 'daeaeFEdbEDaDf'),
		('S_2_1', 'fAAEffcedEafdDeFcDCe'),  # Dividing by total steps is bad for this one.
		('S_2_1', 'eeDcbeBbcdFfDBaDfDeF'),  # Really slow. 4 iterations needed.
		('S_2_1', 'fCbaAdDafeEdbcaabABb'),
		('S_2_1', 'baccabebededdccceeba'),
		('S_2_1', 'FacBcDBACfbDAbCAfEdb'),
		('S_2_1', 'bbdcecbcFA'),  # Reducible with 2 pseudo-Anosov components which are swapped.
		('S_2_1', 'FEFdFCBA'),
		('S_2_1', 'acdCccbBcf'),
		('S_2_1', 'ACBBaF'),  # !?!
		('S_2_1', 'ECdEEEaEce'),
		('S_2_1', 'DCDfCaEd'),
		('S_2_1', 'BaEcCCeAbC'),
		('S_3_1', 'abcdeGh'),  # Takes ~11 iterations to converge!
		('S_3_1', 'fAfdBh'),
		('S_3_1', 'FECC'),
		('S_3_1', 'FdbEBABaGa'),  # Reducible with 2 pA components.
		('S_3_1', 'FdbEABAaGa'),  # These have there invariant laminations in the same cell!
		('S_3_1', 'gEBGhCDHbdgF'),  # The invariant lamination does not have the top eigenvalue.
		('S_3_1', 'edbcAdgGhdHf'),  # Some of these cells only have complex eigenvalues.
		('S_1_2', 'axCxaCACABCcBXbxabCAACACxCCXXXacXCCCXcac'),  # Use to take a long time thanks to overestimated bounds.
		('S_3_1', 'fEbDGFAdagBhAdceCfeE'),  # This has 1 dim eigenspaces that don't lie in the cone.
		]
	
	for index, (surface, word) in enumerate(examples):
		print('%3d/%d: %s %s' % (index+1, len(examples), surface, word), end='')
		S = flipper.load(surface)
		mapping_class = S.mapping_class(word)
		start_time = time()
		try:
			mapping_class.invariant_lamination()
		except flipper.AssumptionError:
			print(', Claim: not pA', end='')
		times[word] = time() - start_time
		print(', Time: %0.3f' % times[word])
	print('Average time: %0.3f' % (sum(times.values()) / len(examples)))
	print('Slowest: %s, Time: %0.3f' % (max(times, key=lambda w: times[w]), max(times.values())))
	print('Total time: %0.3f' % sum(times.values()))

if __name__ == '__main__':
	import argparse
	parser = argparse.ArgumentParser(description='Some of the hardest examples to find invariant laminations for.')
	parser.add_argument('--show', action='store_true', default=False, help='show the source code of this example and exit')
	args = parser.parse_args()
	
	if args.show:
		print(open(__file__, 'r').read())
	else:
		main()






from __future__ import print_function

import os

def main():
	print('Available examples:')
	for example_name in sorted(os.listdir(os.path.dirname(__file__))):
		name, extension = os.path.splitext(example_name)
		if extension == '.py' and not name.startswith('_'):
			print('\tflipper.example.%s' % name)
	print('The --show flag can be used to see an examples source code.')

if __name__ == '__main__':
	main()





''' A module for representing and manipulating matrices.

Provides one class: Matrix.

There are also helper functions: id_matrix and zero_matrix. '''

import flipper
from fractions import Fraction

import itertools

def nonnegative(v):
	''' Return if the given vector is non-negative. '''
	
	return all(x >= 0 for x in v)

def dot(a, b):
	''' Return the dot product of the two given iterables. '''
	
	# Is it significantly faster to do:
	#c = 0
	#for x, y in zip(a, b):
	#	c += x * y
	#return c
	return sum(x * y for x, y in zip(a, b))

def round_fraction(x):
	''' Return the integer closest to this fraction.
	
	Note that we can't use int(round(x)) as x may be too large
	to be represented as a float. '''
	
	# There are three possibilities for what could be the closest integer.
	return min([int(x), int(x) - 1, int(x) + 1], key=lambda y: abs(y - x))

class Matrix(object):
	''' This represents a matrix. '''
	def __init__(self, data):
		assert(isinstance(data, (list, tuple)))
		assert(all(isinstance(row, (list, tuple)) for row in data))
		
		self.rows = [list(row) for row in data]
		self.height = len(self.rows)
		self.width = len(self.rows[0]) if self.height > 0 else 0
		assert(all(len(row) == self.width for row in self))
	def copy(self):
		''' Return a copy of this Matrix. '''
		return Matrix([list(row) for row in self])
	def __getitem__(self, index):
		return self.rows[index]
	def __repr__(self):
		return str(self)
	def __str__(self):
		return '[\n' + ',\n'.join('[' + ', '.join(str(entry) for entry in row) + ']' for row in self) + '\n]'
		# Suggestion: Make Matrices easier to read by omitting zeros.
		#return '[\n' + ',\n'.join('[' + ', '.join(str(entry) if entry != 0 else ' ' for entry in row) + ']' for row in self) + '\n]'
	def __hash__(self):
		return hash(tuple([self.width, self.height] + [tuple(row) for row in self]))
	def __len__(self):
		return self.height
	def __iter__(self):
		return iter(self.rows)
	def __eq__(self, other):
		return self.width == other.width and self.height == other.height and all(row1 == row2 for row1, row2 in zip(self.rows, other.rows))
	def __ne__(self, other):
		return not (self == other)
	
	def __neg__(self):
		return Matrix([[-x for x in row] for row in self])
	def __add__(self, other):
		if isinstance(other, Matrix):
			assert(self.width == other.width and self.height == other.height)
			return Matrix([[a+b for a, b in zip(r1, r2)] for r1, r2 in zip(self, other)])
		else:
			return self + (id_matrix(self.width) * other)
	def __radd__(self, other):
		return self + other
	def __sub__(self, other):
		if isinstance(other, Matrix):
			assert(self.width == other.width and self.height == other.height)
			return Matrix([[self[i][j] - other[i][j] for j in range(self.width)] for i in range(self.height)])
		else:
			return self - (id_matrix(self.width) * other)
	def __rsub__(self, other):
		return -(self - other)
	
	def __call__(self, other):
		assert(isinstance(other, (list, tuple)))
		assert(self.width == 0 or self.width == len(other))
		return [dot(row, other) for row in self]
	
	def __mul__(self, other):
		if isinstance(other, Matrix):
			assert(self.width == 0 or self.width == len(other))
			other_transpose = other.transpose()
			return Matrix([[dot(a, b) for b in other_transpose] for a in self])
		else:  # Multiply entry wise.
			return Matrix([[entry * other for entry in row] for row in self])
	def __rmul__(self, other):
		return self * other
	def __pow__(self, power):
		assert(self.is_square())
		
		if power == 0:
			return id_matrix(self.width)
		elif power == 1:
			return self
		elif power > 1:
			sqrt = self**(power//2)
			square = sqrt * sqrt
			if power % 2 == 1:
				return self * square
			else:
				return square
		else:  # power < 0.
			raise TypeError('Can only raise matrices to positive powers.')
	
	def is_square(self):
		''' Return if this matrix is square. '''
		
		return self.width == self.height
	
	def powers(self, max_power):
		''' Return the list [self**0, ..., self**max_power].
		
		This matrix must be square. '''
		
		assert(self.is_square())
		
		Ms = [id_matrix(self.width)]
		for _ in range(max_power):
			Ms.append(self * Ms[-1])
		return Ms
	def inverse(self):
		''' Return the inverse of this matrix.
		
		This matrix must be square and have determinant +/-1. '''
		
		assert(self.is_square())
		
		P = self.characteristic_polynomial()
		assert(abs(P[0]) == 1)  # Check that the determinant is +/-1.
		# In the next line we should do x // -P[0], but as we know that P[0] == +/-1
		# 1/-P[0] == -P[0] and so we can use a cheaper multiplication.
		return -P[0] * P.shift(-1)(self)
	
	def flatten(self):
		''' Return the entries of this matrix as a single flattened list. '''
		return [entry for row in self for entry in row]
	def transpose(self):
		''' Return the transpose of this matrix. '''
		
		return Matrix(list(zip(*self.rows)))
	def join(self, other):
		''' Return the matrix:
			(self )
			(-----)
			(other)
		This is the same as Sages Matrix.stack() function. '''
		
		return Matrix(self.rows + other.rows)
	def tweak(self, increment, decrement):
		''' Return a copy of this matrix where each increment entry has been increased by 1 and each decrement entry has been decreased by 1. '''
		
		rows = [list(row) for row in self]
		for (i, j) in increment:
			rows[i][j] += 1
		for (i, j) in decrement:
			rows[i][j] -= 1
		return Matrix(rows)
	
	def trace(self):
		''' Return the trace of this matrix. '''
		
		return sum(self[i][i] for i in range(self.width))
	def determinant(self):
		''' Return the determinant of this matrix.
		
		Uses Bareiss' algorithm to compute the determinant in ~O(n^3). See:
			http://cs.nyu.edu/exact/core/download/core_v1.4/core_v1.4/progs/bareiss/bareiss.cpp
		
		This matrix must be square. '''
		
		# We could also just get the constant term of the characteristic polynomial, but this is ~10x faster.
		
		assert(self.is_square())
		
		scale = 1
		A = [list(row) for row in self]
		for i in range(self.width-1):
			if not A[i][i]:  # == 0.
				for j in range(i+1, self.height):
					if A[j][i]:  # != 0.
						A[i], A[j] = A[j], A[i]
						scale = -scale
						break
				else:
					return 0  # We have a column of all 0's.
			for j in range(i+1, self.width):
				for k in range(i+1, self.width):
					A[j][k] = A[j][k]*A[i][i] - A[j][i]*A[i][k]
					if i: A[j][k] = A[j][k] // A[i-1][i-1]  # Division is exact.
		
		return scale * A[self.width-1][self.width-1]
	def characteristic_polynomial(self):
		''' Return the characteristic polynomial of this matrix.
		
		Based off of the Faddeev-Leverrier method. See:
			http://mathfaculty.fullerton.edu/mathews/n2003/FaddeevLeverrierMod.html
		This avoids have to compute an expensive determinant.
		
		This matrix must be square. '''
		
		assert(self.is_square())
		
		# We will actually compute det(\lambdaI - self). Then at the
		# end we correct this by multiplying by the required +/-1.
		A = self
		p = [1] * (self.width+1)
		for i in range(1, self.width+1):
			p[i] = -A.trace() // i
			# If we were smarter we would skip this on the final iteration.
			A = self * (A + p[i])
		# Actually now A / p[i] == A^{-1}.
		sign = +1 if self.width % 2 == 0 else -1
		return flipper.kernel.Polynomial(p[::-1]) * sign
	
	def elementary(self, i, j, k=1):
		''' Return the matrix obtained by performing the elementary move:
			replace row i by row i + k * row j. '''
		
		return Matrix([self[n] if n != i else [x+k*y for x, y in zip(self[i], self[j])] for n in range(self.height)])
	def swap(self, i, j):
		''' Return the matrix obtained by swapping rows i and j. '''
		
		return Matrix([self[j if k == i else i if k == j else k] for k in range(self.height)])
	def row_reduce(self, zeroing_width=None):
		''' Return this matrix after applying elementary row operations
		so that in each row each non-zero entry either:
			1) has a non-zero entry to the left of it,
			2) has no non-zero entries below it, or
			3) is in a column > zeroing_width.
		if zeroing_width is None: zeroing_width = self.width. '''
		
		if zeroing_width is None: zeroing_width = self.width
		i, j = 0, 0
		A = [list(row) for row in self]
		while j < zeroing_width:
			for b, a in itertools.product(range(j, zeroing_width), range(i, self.height)):
				if A[a][b] != 0:
					A[i], A[a] = A[a], A[i]
					j = b
					break
			else:
				break
			
			rlead = A[i][j]
			for k in range(self.height):
				if k != i:
					r2lead = A[k][j]
					if r2lead != 0:
						for x in range(j, self.width):
							A[k][x] = (A[k][x] * rlead) - (A[i][x] * r2lead)
			
			i += 1
			j += 1
		return Matrix(A)
	def kernel(self):
		''' Return a matrix whose rows are a basis for the kernel of this matrix. '''
		
		A = self.join(id_matrix(self.width))
		B = A.transpose()
		C = B.row_reduce(zeroing_width=self.height)
		return Matrix([row[self.height:] for row in C if any(row) and not any(row[:self.height])])
	
	def nonnegative_image(self, v):
		''' Return if self * v >= 0. '''
		
		return all(dot(row, v) >= 0 for row in self)
	
	def simplify(self):
		''' Return a simpler matrix defining the same cone. '''
		
		# Remove nonnegative rows.
		rows = [row for row in self if any(entry < 0 for entry in row)]
		# Remove dupliates.
		rows = list(set(tuple(row) for row in rows))
		# Remove dominated rows.
		rows = [row for row in rows if all(row == row2 or any(x < y for x, y in zip(row, row2)) for row2 in rows)]
		return Matrix(rows)
	
	def solve(self, vector):
		''' Return a vector v such that self * v // det(self) == vector.
		
		Assumes and checks that this matrix is invertible.
		
		This matrix must be square. '''
		
		assert(self.is_square())
		
		if self.determinant() == 0:
			raise flipper.AssumptionError('Matrix is not invertible.')
		
		# This uses the Cayley-Hamilton Theorem that:
		#   self.characteristic_polynomial()(self) == 0
		# Together with the fact that:
		#   self.characteristic_polynomial()(0) == det(self)
		return (-self.characteristic_polynomial().shift(-1)(self))(vector)
	
	def LLL(self, delta=Fraction(3, 4)):
		''' Return a delta-LLL reduced basis for the lattice defined by the rows of this matrix. '''
		
		# Here we will use Fractions to ensure that we retain absolute precision.
		
		# For further information about how this algorithm works see:
		#	* 'Factoring polynomials with rational coefficients' (Lenstra et al. 1982) in particular
		#	  Figure 1, available at http://www.cs.elte.hu/~lovasz/scans/lll.pdf
		#	* LibLLL available at https://github.com/kutio/liblll
		
		m, n = self.width, self.height
		
		M = self
		GS = self  # The Gram-Schmidt orthogonalisation of self.
		mu = zero_matrix(m, n)  # The Gram-Schmidt coefficients.
		for i in range(n):
			for j in range(i):
				mu[i][j] = Fraction(dot(self[i], GS[j]), dot(GS[j], GS[j]))
				GS = GS.elementary(i, j, -mu[i][j])
		B = [dot(GS[i], GS[i]) for i in range(n)]
		
		N = 1
		while True:
			if abs(mu[N][N-1]) > 0.5:
				r = round_fraction(mu[N][N-1])
				M = M.elementary(N, N-1, -r)
				for k in range(N-1):
					mu[N][k] = mu[N][k] - r * mu[N-1][k]
				mu[N][N-1] = mu[N][N-1] - r
			
			if B[N] < (delta - mu[N][N-1] * mu[N][N-1])*B[N-1]:
				u = mu[N][N-1]
				big_B = B[N] + (u * u) * B[N-1]
				mu[N][N-1] = u * Fraction(B[N-1], big_B)
				B[N] = B[N-1] * Fraction(B[N], big_B)
				B[N-1] = big_B
				M = M.swap(N, N-1)
				
				for j in range(N-1):
					mu[N-1][j], mu[N][j] = mu[N][j], mu[N-1][j]
				for i in range(N+1, n):
					mu[i][N-1], mu[i][N] = mu[N][N-1]*mu[i][N-1] + mu[i][N] - u*mu[i][N]*mu[N][N-1], mu[i][N-1] - u * mu[i][N]
				
				if N > 1: N -= 1
			else:
				for j in range(N-2, -1, -1):
					if abs(mu[N][j]) > 0.5:
						r = round_fraction(mu[N][j])
						M = M.elementary(N, j, -r)
						for k in range(j):
							mu[N][k] = mu[N][k] - r * mu[j][k]
						mu[N][j] = mu[N][j] - r
				
				if N == n-1: break
				
				N += 1
		
		return M
	
	def discard_column(self, column):
		''' Return the matrix obtained by discarding the given column. '''
		
		return Matrix([[row[i] for i in range(self.width) if i != column] for row in self])
	
	def find_vector_with_nonnegative_image(self):
		''' Return a vector v != 0 such that self * v >= 0.
		
		Assumes that such a vector exists. '''
		
		# We repeatedly use Fourier--Motzkin elimination to eliminate variables and so
		# reduce the size of the matrix involved. We then reverse this process to construct
		# the required vector.
		# Currently we only use the standard algorithm and so this can result in doubly
		# exponential complexity. See:
		#   http://en.wikipedia.org/wiki/Fourier%E2%80%93Motzkin_elimination
		# for more information.
		
		M = self
		B = []  # Instructions for how to rebuild the vector.
		
		while M.width > 1:
			# Find a good index to reduce by. Ideally, we want one where (almost) all
			# entries of that column have the same sign.
			bias = [0] * M.width
			for row in M:
				for index, entry in enumerate(row):
					bias[index] += 1 if entry > 0 else -1 if entry < 0 else 0
			# The column we will reduce:
			index = max(range(M.width), key=lambda i: abs(bias[i]))
			
			# Perform Fourier--Motzkin elimination with respect to this index.
			pos_rows = [row for row in M if row[index] > 0]
			neg_rows = [row for row in M if row[index] < 0]
			non_rows = [row for row in M if row[index] == 0]
			if len(pos_rows) == 0:  # Problem is independent of x_index.
				M = M.discard_column(index)
				B.append((index, -1))
			elif len(neg_rows) == 0:  # Problem is independent of x_index.
				M = M.discard_column(index)
				B.append((index, 1))
			else:
				new_rows = [[r1[index] * y - r2[index] * x for x, y in zip(r1, r2)] for r1, r2 in itertools.product(pos_rows, neg_rows)]
				M = Matrix(new_rows + non_rows).discard_column(index)
				B.append((index, Matrix(pos_rows)))
		
		# Now that M.width == 1 it is easy to solve.
		if all(entry >= 0 for row in M for entry in row):
			v = [1]
		elif all(entry <= 0 for row in M for entry in row):
			v = [-1]
		else:
			v = [0]
		
		# Reverse the process to rebuild the vector.
		for index, BB in reversed(B):
			if isinstance(BB, Matrix):
				ratios = zip((-BB).discard_column(index)(v), BB.transpose()[index])
				# In this list of ratios all denominators are positive.
				# Get the pair (a, b) with the biggest ratio:
				a, b = ratios[0]
				for p, q in ratios:
					if p * b > a * q:  # If p/q > a/b:
						a, b = p, q
				
				x, scale = a, b
			else:
				x, scale = BB, 1
			v = [entry * scale for entry in v[:index]] + [x] + [entry * scale for entry in v[index:]]
		
		# Check that the vector we've built is nontrivial.
		if all(entry == 0 for entry in v):
			raise flipper.AssumptionError('Polytope is trivial.')
		
		assert(self.nonnegative_image(v))
		
		return v


################################################_
#### Some helper functions for building matrices.

def id_matrix(dim):
	''' Return the identity matrix of given dimension. '''
	
	return Matrix([[1 if i == j else 0 for j in range(dim)] for i in range(dim)])

def zero_matrix(width, height=None):
	''' Return the zero matrix of given dimensions. '''
	
	if height is None: height = width
	return Matrix([[0] * width for _ in range(height)])





''' Stores the documentation for flipper. '''

import os
import sys
import subprocess

def open_documentation(verbose=False):
	''' Open the flipper users guide using the default pdf viewer. '''
	
	datadir = os.path.dirname(__file__)
	disk_file = os.path.join(datadir, 'flipper.pdf')
	if verbose: print('Opening:')
	if verbose: print(disk_file)
	
	# Note that the command contains no user provided data.
	# So even setting shell=True should be safe.
	if sys.platform.startswith('darwin'):
		subprocess.call(['open', disk_file])
	elif sys.platform.startswith('win'):
		os.startfile(disk_file)
	else:
		subprocess.call(['xdg-open', disk_file])





''' A module for representing a triangulation with a preferred flat structure.

Provides two classes: Vector2 and FlatStructure.
	An Vector2 is a vector in RR^2
	A FlatStructure is a triangulation with an assignment of vectors to each edge. '''

import flipper

class Vector2(object):
	''' This represents a point in RR^2. '''
	
	# Warning: This needs to be updated if the interals of this class ever change.
	__slots__ = ['x', 'y']
	
	def __init__(self, x, y):
		self.x, self.y = x, y
	def __str__(self):
		return "(%s, %s)" % (self.x, self.y)
	def __repr__(self):
		return str(self)
	def __iter__(self):
		return iter([self.x, self.y])
	def __eq__(self, other):
		return self.x == other.x and self.y == other.y
	def __ne__(self, other):
		return not (self == other)
	def __neg__(self):
		return Vector2(-self.x, -self.y)
	def __add__(self, other):
		return Vector2(self.x + other.x, self.y + other.y)
	def __sub__(self, other):
		return self + -other
	def __mul__(self, other):
		return Vector2(self.x * other, self.y * other)

class FlatStructure(object):
	''' This represents a triangulation with a flat structure.
	
	It is specified by a triangulation together with a map taking each edge to a Vector2.
	These should satisfy some standard relations like:
		vector[~edge] = -vector[edge], and
		sum(vector[edge] for edge in triangle) == 0.
	These vectors describe the flat structure of self.canonical() and can be used to build a flat surface for self.canonical().
	This is based off of code supplied by Shannon Horrigan.
	'''
	def __init__(self, triangulation, edge_vectors):
		assert(isinstance(triangulation, flipper.kernel.Triangulation))
		assert(isinstance(edge_vectors, dict))
		assert(all(edge in edge_vectors for edge in triangulation.edges))
		assert(all(isinstance(edge_vectors[edge], Vector2) for edge in triangulation.edges))
		
		assert(all(edge_vectors[~edge] == -edge_vectors[edge] for edge in triangulation.edges))
		assert(all(sum([edge_vectors[edge] for edge in triangle], Vector2(0, 0)) == Vector2(0, 0) for triangle in triangulation))
		
		self.triangulation = triangulation
		self.edge_vectors = edge_vectors
	def __str__(self):
		return '\n'.join('%s --> %s' % (edge, self.edge_vectors[edge]) for edge in sorted(self.triangulation.positive_edges, key=lambda e: e.index))
	def __repr__(self):
		return str(self)




''' A module for representing a triangulation of a punctured surface.

Provides five classes: Vertex, Edge, Triangle, Corner and Triangulation.
	A Vertex is a singleton.
	An Edge is an ordered pair of Vertices.
	A Triangle is an ordered triple of Edges.
	A Corner is a Triangle with a chosen side.
	A Triangulation is a collection of Triangles. '''

import flipper

try:
	from Queue import Queue
except ImportError:
	from queue import Queue
from random import choice
from itertools import groupby
import string
from math import log

INFTY = float('inf')

def norm(value):
	''' A map taking an edges label to its index.
	
	That is, x and ~x should map to the same thing. '''
	
	return max(value, ~value)

class Vertex(object):
	''' This represents a vertex, labelled with an integer. '''
	
	# Warning: This needs to be updated if the interals of this class ever change.
	__slots__ = ['label', 'filled']
	
	def __init__(self, label, filled=False):
		assert(isinstance(label, flipper.IntegerType))
		assert(isinstance(filled, bool))
		self.label = label
		self.filled = filled
	
	def __repr__(self):
		return str(self)
	def __str__(self):
		return ('Singularity ' if self.filled else 'Puncture ') + str(self.label)
	def __hash__(self):
		return hash(self.label)
	def __eq__(self, other):
		if isinstance(other, Vertex):
			return self.label == other.label
		else:
			return NotImplemented

class Edge(object):
	''' This represents an oriented edge, labelled with an integer.
	
	It is specified by the vertices that it connects from / to.
	Its inverse edge is created automatically and is labelled with ~its label. '''
	
	# Warning: This needs to be updated if the interals of this class ever change.
	__slots__ = ['source_vertex', 'target_vertex', 'label', 'index', 'reversed_edge']
	
	def __init__(self, source_vertex, target_vertex, label, reversed_edge=None):
		assert(isinstance(source_vertex, Vertex))
		assert(isinstance(target_vertex, Vertex))
		assert(isinstance(label, flipper.IntegerType))
		assert(reversed_edge is None or isinstance(reversed_edge, Edge))
		
		self.source_vertex = source_vertex
		self.target_vertex = target_vertex
		self.label = label
		self.index = norm(self.label)
		if reversed_edge is None: reversed_edge = Edge(self.target_vertex, self.source_vertex, ~self.label, self)
		self.reversed_edge = reversed_edge
	
	def __repr__(self):
		return str(self)
	def __str__(self):
		return ('' if self.is_positive() else '~') + str(self.index)
	def __hash__(self):
		return hash(self.label)
	def __eq__(self, other):
		if isinstance(other, Edge):
			return self.label == other.label
		else:
			return NotImplemented
	
	def __invert__(self):
		''' Return this edge but with reversed orientation. '''
		
		return self.reversed_edge
	
	def is_positive(self):
		''' Return if this edge is the positively oriented one. '''
		
		return self.label == self.index
	
	def sign(self):
		''' Return the sign (+/-1) of this edge. '''
		
		return +1 if self.is_positive() else -1

class Triangle(object):
	''' This represents a triangle.
	
	It is specified by a list of three edges, ordered anticlockwise.
	It builds its corners automatically. '''
	
	# Warning: This needs to be updated if the interals of this class ever change.
	__slots__ = ['edges', 'labels', 'indices', 'vertices', 'corners']
	
	def __init__(self, edges):
		assert(isinstance(edges, (list, tuple)))
		assert(all(isinstance(edge, Edge) for edge in edges))
		assert(len(edges) == 3)
		
		# Edges are ordered anti-clockwise. We will cyclically permute
		# these to a canonical ordering, the one where the edges are ordered
		# minimally by label.
		best_index = min(range(3), key=lambda i: edges[i].label)
		
		self.edges = edges[best_index:] + edges[:best_index]
		self.labels = [edge.label for edge in self]
		self.indices = [edge.index for edge in self]
		self.vertices = [self.edges[1].target_vertex, self.edges[2].target_vertex, self.edges[0].target_vertex]
		self.corners = [Corner(self, i) for i in range(3)]
	
	def __repr__(self):
		return str(self)
	def __str__(self):
		return str(tuple(self.edges))
	def __hash__(self):
		return hash(tuple(self.labels))
	def __eq__(self, other):
		if isinstance(other, Triangle):
			return self.labels == other.labels
		else:
			return NotImplemented
	
	# Note that this is NOT the same convention as used in pieces.
	# There iterating and index accesses return vertices.
	def __iter__(self):
		return iter(self.edges)
	
	def __getitem__(self, index):
		return self.edges[index]
	
	def __contains__(self, item):
		if isinstance(item, flipper.IntegerType):
			return item in self.labels
		elif isinstance(item, Edge):
			return item in self.edges
		elif isinstance(item, Vertex):
			return item in self.vertices
		elif isinstance(item, Corner):
			return item in self.corners
		else:
			return NotImplemented

class Corner(object):
	''' This represents a corner of a triangulation
	
	It is a triangle along with a side number (the side opposite this corner). '''
	
	# Warning: This needs to be updated if the interals of this class ever change.
	__slots__ = ['triangle', 'side', 'edges', 'labels', 'indices', 'vertices', 'label', 'index', 'vertex', 'edge']
	
	def __init__(self, triangle, side):
		assert(isinstance(triangle, Triangle))
		assert(isinstance(side, flipper.IntegerType))
		
		self.triangle = triangle
		self.side = side
		
		self.edges = self.triangle.edges[self.side:] + self.triangle.edges[:self.side]
		self.labels = self.triangle.labels[self.side:] + self.triangle.labels[:self.side]
		self.indices = self.triangle.indices[self.side:] + self.triangle.indices[:self.side]
		self.vertices = self.triangle.vertices[self.side:] + self.triangle.vertices[:self.side]
		
		self.label = self.labels[0]
		self.index = self.indices[0]
		self.vertex = self.vertices[0]
		self.edge = self.edges[0]
	
	def __repr__(self):
		return str(self)
	def __str__(self):
		return str((self.triangle, self.side))

# Remark: In other places in the code you will often see L(triangulation). This is the space
# of laminations on triangulation with the coordinate system induced by the triangulation.

class Triangulation(object):
	''' This represents a triangulation of a punctured surface.
	
	It is specified by a list of Triangles. Its edges must be
	numbered 0, 1, ... and its vertices must be numbered 0, 1, ... '''
	def __init__(self, triangles):
		assert(isinstance(triangles, (list, tuple)))
		assert(all(isinstance(triangle, Triangle) for triangle in triangles))
		
		# We will sort the triangles into a canonical ordering, the one where the edges are ordered
		# minimally by label. This allows for fast comparisons.
		self.triangles = sorted(triangles, key=lambda t: [e.label for e in t])
		
		self.edges = [edge for triangle in self for edge in triangle.edges]
		self.positive_edges = [edge for edge in self.edges if edge.is_positive()]
		self.labels = sorted([edge.label for edge in self.edges])
		self.indices = sorted([edge.index for edge in self.positive_edges])
		self.vertices = sorted(set(vertex for triangle in self for vertex in triangle.vertices), key=lambda vertex: vertex.label)
		assert(not all(vertex.filled for vertex in self.vertices))
		self.corners = [corner for triangle in self for corner in triangle.corners]
		
		self.num_triangles = len(self.triangles)
		self.zeta = len(self.positive_edges)
		assert(self.zeta == self.num_triangles * 3 // 2)
		self.num_vertices = len(self.vertices)
		self.num_filled_vertices = len([vertex for vertex in self.vertices if vertex.filled])
		self.num_unfilled_vertices = self.num_vertices - self.num_filled_vertices
		# Check that the vertices are labelled 0, ..., num_vertices-1.
		assert(set(vertex.label for vertex in self.vertices) == set(range(self.num_vertices)))
		# Check that the edges have indices 0, ..., zeta-1.
		assert(set(self.labels) == set([i for i in range(self.zeta)] + [~i for i in range(self.zeta)]))
		
		self.triangle_lookup = dict((edge.label, triangle) for triangle in self for edge in triangle.edges)
		self.edge_lookup = dict((edge.label, edge) for edge in self.edges)
		self.corner_lookup = dict((corner.label, corner) for corner in self.corners)
		self.vertex_lookup = dict((corner.label, corner.vertex) for corner in self.corners)
		
		# This appears to be one of the slowest bits when there is a high degree vertex.
		def order_corner_class(corner_class):
			''' Return the given corner_class but reorderd so that corners occur anti-clockwise about the vertex. '''
			
			corner_class = list(corner_class)
			lookup = dict((corner.edges[2], corner) for corner in corner_class)
			ordered_class = [None] * len(corner_class)
			ordered_class[0] = corner_class[0]  # Get one corner to start at.
			# Perhaps this should be chosen in some canonical way. Smallest labelled one?
			# This isn't totally safe: it doesn't check that there aren't multiple cycles.
			for i in range(len(corner_class)-1):
				try:
					ordered_class[i+1] = lookup[~ordered_class[i].edges[1]]
				except KeyError:
					raise ValueError('Corners do not close up about vertex.')
			
			if ordered_class[0].edges[2] != ~ordered_class[-1].edges[1]:
				raise ValueError('Corners do not close up about vertex.')
			
			return ordered_class
		
		# We want to sort by the vertices so groupby will gather them but this would require adding
		# orderings to the Vertex class so we'll just sort / group by the vertex label. This should
		# uniquely determine the vertex.
		vertexer = lambda corner: corner.vertex.label
		self.corner_classes = [order_corner_class(g) for _, g in groupby(sorted(self.corners, key=vertexer), key=vertexer)]
		
		self.euler_characteristic = self.num_filled_vertices - self.zeta + self.num_triangles  # V - E + F.
		# NOTE: This assumes connected.
		self.genus = (2 - self.euler_characteristic - self.num_unfilled_vertices) // 2
		
		# The maximum order of a periodic mapping class.
		# These bounds follow from the 4g + 4 bound on the closed surface [Primer reference]
		# and the Riemann removable singularity theorem which allows us to cap off the
		# punctures when the genus > 1 without affecting this bound.
		# NOTE: This assumes connected.
		if self.genus > 1:
			self.max_order = 4 * self.genus + 2
		elif self.genus == 1:
			self.max_order = max(self.num_unfilled_vertices, 6)
		else:
			self.max_order = self.num_unfilled_vertices
		
		# Two triangualtions are the same if and only if they have the same signature.
		self.signature = [e.label for t in self for e in t]
	
	@classmethod
	def from_tuple(cls, edge_labels, vertex_labels=None, vertex_states=None):
		''' Return an Triangulation from a list of triples of edge labels.
		
		Let T be an ideal triangulaton of the punctured (oriented) surface S. Orient
		and edge e of T and assign an index i(e) in 0, ..., zeta-1. Now to each
		triangle t of T associate the triple j)t) := (j(e_1), j(e_2), j(e_3)) where:
			- e_1, e_2, e_3 are the edges of t, ordered acording to the orientation of t, and
			- j(e) = {  i(e) if the orientation of e agrees with that of t, and
					 { ~i(e) otherwise.
				Here ~x := -1 - x, the two's complement of x.
		
		We may describe T by the list [j(t) for t in T]. This function reconstructs
		T from such a list.
		
		edge_labels must be a list of triples of integers and each of
		0, ..., zeta-1, ~0, ..., ~(zeta-1) must occur exactly once.
		
		Two optional arguments allow the states of vertices to be specified. These
		are intended to only really be used by pickling methods.
		
		Firstly, if given, vertex_labels is a dictionary taking x to the label of the
		vertex opposite edge x. Vertices can be labelled by anything but using something
		like 0, ..., num_vertices-1 is sensible.
		
		Secondly, if given, vertex_states is a dict, list or tuple of Boolean flags such that
		the vertex labelled i is filled iff vertex_states[i] == True. '''
		
		assert(isinstance(edge_labels, (list, tuple)))
		assert(all(isinstance(labels, (list, tuple)) for labels in edge_labels))
		assert(all(len(labels) == 3 for labels in edge_labels))
		assert(vertex_labels is None or isinstance(vertex_labels, dict))
		assert(vertex_states is None or isinstance(vertex_states, (list, tuple, dict)))
		assert(len(edge_labels) > 0)
		
		zeta = len(edge_labels) * 3 // 2
		
		# Check that each of 0, ..., zeta-1, ~0, ..., ~(zeta-1) occurs exactly once.
		flattened = [label for labels in edge_labels for label in labels]
		for i in range(zeta):
			if i not in flattened:
				raise TypeError('Missing label %d' % i)
			if ~i not in flattened:
				raise TypeError('Missing label ~%d' % i)
		
		def finder(edge_label):
			''' Return the label and position of the given edge_label. '''
			
			for labels in edge_labels:
				for i in range(3):
					if labels[i] == edge_label:
						return (labels, i)
			raise flipper.FatalError('Label now missing.')
		
		def rotate(edge_label):
			''' Return the edge label one click round from edge_label. '''
			
			label, i = finder(edge_label)
			return label[(i+1) % 3]
		
		# Group the edges into vertex classes. Here two edges are in the same
		# class iff they have the same tail.
		unused = [i for i in range(zeta)] + [~i for i in range(zeta)]
		vertex_classes = []
		while unused:
			new_vertex = [unused.pop()]
			while True:
				label, i = finder(new_vertex[-1])
				neighbour = ~label[(i+2) % 3]
				if neighbour in unused:
					new_vertex.append(neighbour)
					unused.remove(neighbour)
				else:
					break
			
			vertex_classes.append(new_vertex)
		
		num_vertices = len(vertex_classes)
		
		# Build the vertex_labels if not given.
		if vertex_labels is None:
			vertex_labels = dict()
			# We will label the vertices 0, ..., num_vertices-1 in some order.
			for index, vertex_class in enumerate(vertex_classes):
				for edge_label in vertex_class:
					vertex_labels[rotate(edge_label)] = index
		
		# Sanity check vertex_labels now.
		# All edges should have a label.
		# Edges should have the same label iff they are opposite the same vertex.
		
		for edge_label in [i for i in range(zeta)] + [~i for i in range(zeta)]:
			if edge_label not in vertex_labels:
				raise TypeError('Missing vertex label for edge %d.' % edge_label)
		
		for vertex_class in vertex_classes:
			for edge_label in vertex_class:
				if vertex_labels[rotate(edge_label)] != vertex_labels[rotate(vertex_class[0])]:
					raise TypeError('Edges %d and %d should not have different vertex labels.' % (rotate(edge_label), rotate(vertex_class[0])))
		
		X = set(vertex_labels.values())
		# Check we have the right number of labels. This also checks that distinct vertex_classes have distinct labels.
		if len(X) != num_vertices:
			raise TypeError('There are %d vertices but %d vertex labels were given.' % (num_vertices, len(X)))
		
		# Build the vertex_states if not given.
		if vertex_states is None:
			vertex_states = {label: False for label in X}
		
		# Check we have a state for each vertex.
		for label in X:
			if label not in vertex_states:
				raise TypeError('No state given for vertex %s.' % label)
		
		# Build the vertices.
		vertices = [Vertex(label, filled=vertex_states[label]) for label in X]
		
		def vertexer(edge_label):
			''' Return the vertex at the tail of the given edge label. '''
			
			return vertices[vertex_labels[rotate(edge_label)]]
		
		# Build the Edges.
		edges_map = dict((i, Edge(vertexer(i), vertexer(~i), i)) for i in range(zeta))
		for i in range(zeta):
			edges_map[~i] = ~edges_map[i]
		
		triangles = [Triangle([edges_map[label] for label in labels]) for labels in edge_labels]
		
		return cls(triangles)
	
	@classmethod
	def from_string(cls, signature):
		''' Return the triangulation described by the given isomorphism signature.
		
		See the appendix of:
			Simplification paths in the Pachner graphs of closed orientable
			3-manifold triangulations
		for a more detailed description of this construction. '''
		
		# We will specialise this function for loading isomorphism signatures
		# of closed 2--manifolds only. This will enable us to remove a lot of
		# variables and simplify proceedings.
		
		assert(isinstance(signature, flipper.StringType))
		
		char = string.ascii_lowercase + string.ascii_uppercase + string.digits + '+-'
		char_lookup = dict((letter, index) for index, letter in enumerate(char))
		perm_lookup = flipper.kernel.permutation.all_permutations(3)
		
		def debase(digits, base=64):
			''' Return the decimal corresponding to a base64 sequence of digits. '''
			
			return sum(digit * base**index for index, digit in enumerate(digits))
		
		try:
			values = [char_lookup[letter] for letter in signature]
		except KeyError:
			raise ValueError('Signature must be a string matching [a-zA-Z0-9+-]*')
		
		if values[0] < 63:
			num_chars = 1
			num_tri = values[0]
			start = 1
		else:
			num_chars = values[1]  # This must be > 1.
			num_tri = debase(values[1:num_chars])
			start = 1 + num_chars
		
		if num_chars == 0:
			raise ValueError('Signature must specify a character length > 0.')
			
		if num_tri == 0:
			raise ValueError('Signature must specify at least one triangle.')
		
		start2 = start + num_tri // 2  # Type sequence is [start:start2].
		start3 = start2 + num_chars * (1 + num_tri // 2)  # Destination sequence is [start2:start3].
		start4 = start3 + (1 + num_tri // 2)  # Permutation sequence is [start3:start4].
		if start4 != len(values):
			raise ValueError('Signature must specify permutations for each triangle.')
		
		type_sequence = [t for value in values[start:start2] for t in [value % 4, (value // 4) % 4, (value // 16) % 4]]
		destination_sequence = [debase(values[i:i+num_chars]) for i in range(start2, start3, num_chars)]
		permutation_sequence = [perm_lookup[value] for value in values[start3:start4]]
		
		zeta = 0
		num_tri_used = 1
		type_index, destination_index, permutation_index = 0, 0, 0
		edge_labels = [[None, None, None] for _ in range(num_tri)]
		triangle_reversed = [None] * num_tri
		triangle_reversed[0] = False
		for i in range(num_tri):
			for j in range(3):
				if edge_labels[i][j] is None:  # Otherwise we have filled in this entry from the other side.
					try:
						if type_sequence[type_index] == 0:
							# We cannot yet handle triangulations with boundary.
							raise ValueError('Triangulations must be closed and so gluing must not be type 0.')
						elif type_sequence[type_index] == 1:
							target, gluing = num_tri_used, perm_lookup[0]
							triangle_reversed[target] = not triangle_reversed[i]
							
							num_tri_used += 1
						elif type_sequence[type_index] == 2:
							target, gluing = destination_sequence[destination_index], permutation_sequence[permutation_index]
							
							destination_index += 1
							permutation_index += 1
						else:
							raise ValueError('Each gluing must be type 0, 1 or 2.')
					except IndexError:
						raise ValueError('String does not correspond to a isomorphism signature.')
					type_index += 1
					
					edge_labels[i][j] = zeta
					edge_labels[target][gluing(j)] = ~zeta
					zeta += 1
		
		if num_tri_used != num_tri:
			raise ValueError('Unused triangles. String does not correspond to a isomorphism signature.')
		# Check there are no unglued edges.
		if not all(all(entry is not None for entry in row) for row in edge_labels):
			raise ValueError('Unused edges. String does not correspond to a isomorphism signature.')
		
		edge_labels = [edge_labels[i] if triangle_reversed[i] else edge_labels[i][::-1] for i in range(num_tri)]
		
		return cls.from_tuple(edge_labels)
	
	def __repr__(self):
		return str(self)
	def __str__(self):
		return str(list(self))
	def __iter__(self):
		return iter(self.triangles)
	def __getitem__(self, index):
		return self.triangles[index]
	def __contains__(self, item):
		if isinstance(item, Vertex):
			return item in self.vertices
		elif isinstance(item, Edge):
			return item in self.edges
		elif isinstance(item, Triangle):
			return item in self.triangles
		elif isinstance(item, Corner):
			return item in self.corners
		else:
			return NotImplemented
	def package(self):
		''' Return a small amount of info that create_triagulation can use to reconstruct this triangulation. '''
		return (
			[t.labels for t in self],
			{corner.label: corner.vertex.label for corner in self.corners},
			{vertex.label: vertex.filled for vertex in self.vertices}
			)
	def __reduce__(self):
		# Triangulations are already pickleable but this results in a much smaller pickle.
		return (create_triangulation, (self.__class__,) + self.package())
	def __eq__(self, other):
		return self.signature == other.signature
	def __ne__(self, other):
		return not(self == other)
	
	def vertices_of_edge(self, edge_label):
		''' Return the two vertices at the ends of the given edge. '''
		
		# Refactor out?
		return [self.edge_lookup[edge_label].source_vertex, self.edge_lookup[~edge_label].source_vertex]
	
	def triangles_of_edge(self, edge_label):
		''' Return the two triangles containing the given edge. '''
		
		# Refactor out?
		return [self.triangle_lookup[edge_label], self.triangle_lookup[~edge_label]]
	
	def corner_of_edge(self, edge_label):
		''' Return the corner opposite the given edge. '''
		assert(isinstance(edge_label, flipper.IntegerType))
		
		# Refactor out?
		return self.corner_lookup[edge_label]
	
	def corner_class_of_vertex(self, vertex):
		''' Return the corner class containing the given vertex. '''
		
		# Refactor out?
		for cc in self.corner_classes:
			if cc[0].vertex == vertex:
				return cc
		
		raise ValueError('Given vertex does not correspond to a corner class.')
	
	def opposite_corner(self, corner):
		''' Return the corner opposite the given corner. '''
		
		return self.corner_of_edge(~(corner.label))
	
	def rotate_corner(self, corner):
		''' Return the corner obtained by rotating the given corner one click anti-clockwise. '''
		
		return self.corner_of_edge(corner.labels[1])
	
	def iso_sig(self, preserve_orientation=False, skip=None, start_points=None):
		''' Return the isomorphism signature of this triangulation as described by Ben Burton.
		
		This is a string such that two triangulations have the same signature
		if and only if there is a homeomorphism taking one to the other.
		
		If skip is not None then it must be an iterable containing the labels
		of edges to treat as boundary. If preserve_orientation is True then
		this identifying homeomorphism must be orientation preserving.
		
		
		If start_points is not None then it must be an iterable containing pairs:
		(edge_lable, oriented) which specifies a corner to start with and whether
		to orient this corner to match the orientation of the triangle. This can
		be used to generate signatures relative to a fixed boundary by specify
		an edge on that boundary (and True) as the only starting point. '''
		
		best = ([INFTY], [INFTY], [INFTY])
		skip = set() if skip is None else set(skip)
		
		perm_inverse = flipper.kernel.permutation.PERM3_INVERSE
		perm_lookup = flipper.kernel.permutation.PERM3_LOOKUP
		transition_perm_lookup = flipper.kernel.permutation.TRANSITION_PERM3_LOOKUP
		perm_reverse = flipper.kernel.Permutation([0, 2, 1])
		
		# Set up the starting points.
		if start_points is None:
			# If we want to preserve the orientation on the surface we should only use the positive
			# permutations. So we should set the orientation to True.
			if preserve_orientation:
				start_points = [(label, True) for label in self.labels]
			else:
				start_points = [(label, orientation) for label in self.labels for orientation in [True, False]]
		
		for start_edge, start_orientation in start_points:
			start_corner = self.corner_lookup[start_edge]
			start_triangle = start_corner.triangle
			
			if not all(label in skip for label in start_triangle.labels):
				start_perm = flipper.kernel.permutation.cyclic_permutation(start_corner.side, 3).inverse()
				if not start_orientation:
					start_perm = start_perm * perm_reverse
				
				type_sequence = []
				target_sequence = []
				permutation_sequence = []
				
				good = True
				queue = Queue()
				queue.put(start_triangle)
				triangle_labels = {start_triangle: (0, start_perm)}
				num_triangles_seen = 1
				
				while not queue.empty() and good:
					triangle = queue.get()
					_, perm = triangle_labels[triangle]
					perm_inv = perm_inverse[perm]
					
					for j in range(3):
						side = perm_inv(j)
						target_corner = self.corner_of_edge(~triangle.labels[side])
						target_triangle = target_corner.triangle
						target_side = target_corner.side
						if ~triangle.labels[side] in skip:
							# This edge was really a boundary edge.
							type_sequence.append(0)
						elif target_triangle not in triangle_labels:
							target_perm = perm * transition_perm_lookup[(target_side, side)]
							triangle_labels[target_triangle] = (num_triangles_seen, target_perm)
							queue.put(target_triangle)
							num_triangles_seen += 1
							
							type_sequence.append(1)
							# We don't need to record the follow as they are implied.
							# target_sequence.append(len(queue))
							# permutation_sequence.append(perm_lookup[id_perm])
						else:
							triangle_index, _ = triangle_labels[triangle]
							target_index, target_perm = triangle_labels[target_triangle]
							k = target_perm(target_side)
							if target_index > triangle_index or (target_index == triangle_index and k > j):
								# We've not done this gluing yet.
								transition_perm = target_perm * transition_perm_lookup[(side, target_side)] * perm_inv
								
								type_sequence.append(2)
								target_sequence.append(target_index)
								permutation_sequence.append(perm_lookup[transition_perm])
						# We can give up early if we've built something bigger than best.
						if type_sequence > best[0]:
							good = False
							break
				
				if good:
					best = min((type_sequence, target_sequence, permutation_sequence), best)
		
		char = string.ascii_lowercase + string.ascii_uppercase + string.digits + '+-'
		
		type_sequence, target_sequence, permutation_sequence = best
		# Pad the type_sequence with 0's so that its length is a multiple of 3.
		type_sequence = type_sequence + [0] * (-len(type_sequence) % 3)
		char_type = ''.join(char[type_sequence[i] + 4 * type_sequence[i+1] + 16 * type_sequence[i+2]] for i in range(0, len(type_sequence), 3))
		char_perm = ''.join(char[perm] for perm in permutation_sequence)
		
		# Get the number of triangles that we encoutered.
		num_tri = type_sequence.count(1) + 1
		
		if num_tri < 63:
			char_start = char[num_tri]
			char_target = ''.join(char[target] for target in target_sequence)
		else:
			digits = int(log(num_tri) / log(64)) + 1
			char_start = char[63] + char[digits] + ''.join(char[(num_tri // 64**i) % 64] for i in range(digits))
			char_target = ''.join(char[(target // 64**i) % 64] for target in target_sequence for i in range(digits))
		
		return char_start + char_type + char_target  + char_perm
	
	def is_flippable(self, edge_label):
		''' Return if the given edge is flippable.
		
		An edge is flippable if and only if it lies in two distinct triangles. '''
		
		return self.triangle_lookup[edge_label] != self.triangle_lookup[~edge_label]
	
	def flippable_edges(self):
		''' Return this list of flippable edges of this triangulation. '''
		
		return [i for i in self.indices if self.is_flippable(i)]
	
	def nonflippable_boundary(self, edge_label):
		''' Return the label of the edge bounding the once-punctured monogon containing edge_label.
		
		The given edge must not be flippable. '''
		
		assert(not self.is_flippable(edge_label))
		
		# As edge_label is not flippable the triangle containing it must be (edge_label, ~edge_label, x).
		[boundary_edge] = [label for label in self.triangle_lookup[edge_label].labels if label != edge_label and label != ~edge_label]
		return boundary_edge
	
	def square_about_edge(self, edge_label):
		''' Return the four edges around the given edge.
		
		The chosen edge must be flippable. '''
		
		assert(self.is_flippable(edge_label))
		
		# Given the label e, return the edges a, b, c, d in order.
		#
		# #<----------#
		# |     a    ^^
		# |         / |
		# |  A     /  |
		# |       /   |
		# |b    e/   d|
		# |     /     |
		# |    /      |
		# |   /       |
		# |  /     B  |
		# | /         |
		# V/    c     |
		# #---------->#
		
		corner_A, corner_B = self.corner_of_edge(edge_label), self.corner_of_edge(~edge_label)
		return [corner_A.edges[1], corner_A.edges[2], corner_B.edges[1], corner_B.edges[2]]
	
	def flip_edge(self, edge_label):
		''' Return a new triangulation obtained by flipping the given edge.
		
		The chosen edge must be flippable. '''
		
		assert(self.is_flippable(edge_label))
		
		# Use the following for reference:
		# #<----------#     #-----------#
		# |     a    ^^     |\          |
		# |         / |     | \         |
		# |  A     /  |     |  \     A2 |
		# |       /   |     |   \       |
		# |b    e/   d| --> |    \e'    |
		# |     /     |     |     \     |
		# |    /      |     |      \    |
		# |   /       |     |       \   |
		# |  /     B  |     | B2     \  |
		# | /         |     |         \ |
		# V/    c     |     |          V|
		# #---------->#     #-----------#
		
		vertex_map = dict()
		for vertex in self.vertices:
			vertex_map[vertex] = flipper.kernel.Vertex(vertex.label, filled=vertex.filled)
		
		edge_map = dict()
		# Far away edges should go to an exact copy of themselves.
		for edge in self.positive_edges:
			edge_map[edge] = flipper.kernel.Edge(vertex_map[edge.source_vertex], vertex_map[edge.target_vertex], edge.label)
			edge_map[~edge] = ~edge_map[edge]
		
		a, b, c, d = self.square_about_edge(edge_label)
		# We need to label new_edge with norm(edge_label) so that self.flip_edge(i).flip_edge(~i) == self.
		new_edge = Edge(vertex_map[a.target_vertex], vertex_map[c.target_vertex], norm(edge_label))
		
		triangle_A2 = Triangle([new_edge, edge_map[d], edge_map[a]])
		triangle_B2 = Triangle([~new_edge, edge_map[b], edge_map[c]])
		
		triangles = [flipper.kernel.Triangle([edge_map[edge] for edge in triangle]) for triangle in self if edge_label not in triangle.labels and ~edge_label not in triangle]
		
		return Triangulation(triangles + [triangle_A2, triangle_B2])
	
	def relabel_edges(self, label_map):
		''' Return a new triangulation obtained by relabelling the edges according to label_map. '''
		
		if isinstance(label_map, (list, tuple)):
			label_map = dict(enumerate(label_map))
		else:
			label_map = dict(label_map)
		
		# Build any missing labels.
		for i in self.indices:
			if i in label_map and ~i in label_map:
				pass
			elif i not in label_map and ~i in label_map:
				label_map[i] = ~label_map[~i]
			elif i in label_map and ~i not in label_map:
				label_map[~i] = ~label_map[i]
			else:
				raise flipper.AssumptionError('Missing new label for %d.' % i)
		
		vertex_map = dict()
		for vertex in self.vertices:
			vertex_map[vertex] = flipper.kernel.Vertex(vertex.label, filled=vertex.filled)
		
		edge_map = dict()
		# Far away edges should go to an exact copy of themselves.
		for edge in self.positive_edges:
			edge_map[edge] = flipper.kernel.Edge(vertex_map[edge.source_vertex], vertex_map[edge.target_vertex], label_map[edge.label])
			edge_map[~edge] = ~edge_map[edge]
		
		triangles = [flipper.kernel.Triangle([edge_map[edge] for edge in triangle]) for triangle in self]
		return Triangulation(triangles)
	
	def tree_and_dual_tree(self, respect_fillings=False):
		''' Return a maximal tree in the 1--skeleton of this triangulation and a
		maximal tree in 1--skeleton of the dual of this triangulation.
		
		These are given as lists of Booleans signaling if each edge is in the tree.
		No edge is used in both the tree and the dual tree. Note that when this surface
		is disconnected this tree is actually a forest. '''
		
		components = self.components()
		
		tree = [False] * self.zeta
		vertices_used = dict((vertex, False) for vertex in self.vertices)
		# Get some starting vertices.
		if respect_fillings:
			for vertex in self.vertices:
				if not vertex.filled:
					vertices_used[vertex] = True
		else:
			for component in components:
				for edge_label in component:
					vertex = self.edge_lookup[edge_label].source_vertex
					if not vertex.filled:
						vertices_used[vertex] = True
						break
		
		if False:
			for vertex in self.vertices:
				if not vertex.filled:
					vertices_used[vertex] = True
					if not respect_fillings:
						# Stop as soon as we've marked one.
						break
		
		while True:
			for edge_index in range(self.zeta):
				if not tree[edge_index]:
					a, b = self.vertices_of_edge(edge_index)
					if vertices_used[a] != vertices_used[b]:
						tree[edge_index] = True
						vertices_used[a] = True
						vertices_used[b] = True
						break
			else:
				break  # If there are no more to add then our tree is maximal.
		
		dual_tree = [False] * self.zeta
		faces_used = dict((triangle, False) for triangle in self.triangles)
		for component in components:
			faces_used[self.triangle_lookup[component[0]]] = True
		
		while True:
			for edge_index in range(self.zeta):
				if not tree[edge_index] and not dual_tree[edge_index]:
					a, b = self.triangles_of_edge(edge_index)
					if faces_used[a] != faces_used[b]:
						dual_tree[edge_index] = True
						faces_used[a] = True
						faces_used[b] = True
						break
			else:
				break  # If there are no more to add then our dual tree is maximal.
		
		return tree, dual_tree
	
	def homology_basis(self):
		''' Return a basis for H_1 of the underlying punctured surface.
		
		Each element is given as a path in the dual 1--skeleton and corrsponds
		to a good curve. Each path will meet each edge at most once.
		
		Each pair of paths is guaranteed to meet at most once. '''
		
		# Construct a maximal spanning tree in the 1--skeleton of the triangulation.
		# and a maximal spanning tree in the complement of the tree in the 1--skeleton of the dual triangulation.
		tree, dual_tree = self.tree_and_dual_tree()
		
		# Generators are given by edges not in the tree or the dual tree (along with some segment
		# in the dual tree to make it into a loop).
		dual_tree_indices = [edge_index for edge_index in range(self.zeta) if dual_tree[edge_index]]
		homology_generators = []
		for edge_index in range(self.zeta):
			if not tree[edge_index] and not dual_tree[edge_index]:
				generator = [~edge_index]
				source, target = self.triangles_of_edge(edge_index)
				
				# Find a path in dual_tree from source to target. This is a really
				# inefficient way to do this. We initially define the distance to
				# each point to be self.num_triangles+1 which we know is larger than
				# any possible distance.
				distance = dict((triangle, self.num_triangles+1) for triangle in self.triangles)
				distance[source] = 0
				while distance[target] == self.num_triangles+1:
					for edge in dual_tree_indices:  # Only allowed to move in the tree.
						a, b = self.triangles_of_edge(edge)
						distance[a] = min(distance[b]+1, distance[a])
						distance[b] = min(distance[a]+1, distance[b])
				
				# Now find a way from the target back to the source by following the distance
				current = target
				while current != source:
					for edge in dual_tree_indices:
						a, b = self.triangles_of_edge(edge)
						if b == current and distance[a] == distance[b]-1:
							generator.append(edge)
							current = a
							break
						elif a == current and distance[b] == distance[a]-1:
							generator.append(~edge)
							current = b
							break
				
				# We've now made a generating loop.
				homology_generators.append(generator)
		
		return homology_generators
	
	def find_isometry(self, other, label_map, respect_fillings=True):
		''' Return the isometry from this triangulation to other defined by label_map.
		
		label_map must be a dictionary mapping self.labels to other.labels. Labels may
		be omitted if they are determined by other given ones and these will be found
		automatically.
		
		Assumes (and checks) that such an isometry exists and is unique. '''
		
		assert(isinstance(label_map, dict))
		
		# Make a local copy as we may need to make a lot of changes.
		label_map = dict(label_map)
		
		source_orders = dict([(corner.label, len(corner_class)) for corner_class in self.corner_classes for corner in corner_class])
		target_orders = dict([(corner.label, len(corner_class)) for corner_class in other.corner_classes for corner in corner_class])
		# We do a depth first search extending the corner map across the triangulation.
		# This is a stack of labels that may still have consequences to check.
		to_process = [(edge_from_label, label_map[edge_from_label]) for edge_from_label in label_map]
		while to_process:
			from_label, to_label = to_process.pop()
			
			neighbours = [
				(~from_label, ~to_label),
				(self.corner_lookup[from_label].labels[1], other.corner_lookup[to_label].labels[1])
				]
			for new_from_label, new_to_label in neighbours:
				if new_from_label in label_map:
					# Check that this map is still consistent.
					if new_to_label != label_map[new_from_label]:
						raise flipper.AssumptionError('This label_map does not extend to an isometry.')
				else:
					# Extend the map.
					if source_orders[new_from_label] != target_orders[new_to_label] or \
						respect_fillings and self.vertex_lookup[new_from_label].filled != other.vertex_lookup[new_to_label].filled:
						raise flipper.AssumptionError('This label_map does not extend to an isometry.')
					label_map[new_from_label] = new_to_label
					to_process.append((new_from_label, new_to_label))
		
		if any(i not in label_map for i in self.labels):
			raise flipper.AssumptionError('This label_map cannot be extended to an isometry.')
		
		return flipper.kernel.Isometry(self, other, label_map)
	
	def isometries_to(self, other, respect_fillings=True):
		''' Return a list of all isometries from this triangulation to other. '''
		
		assert(isinstance(other, Triangulation))
		assert(isinstance(respect_fillings, bool))
		
		if self.zeta != other.zeta:
			return []
		
		# !?! This needs to be modified to work on disconnected surfaces.
		
		# Isometries are determined by where a single triangle is sent.
		# We take a corner of smallest degree.
		source_cc = min(self.corner_classes, key=len)
		source_corner = source_cc[0]
		# And find all the places where it could be sent so there are as few as possible to check.
		target_corners = [corner for target_cc in other.corner_classes for corner in target_cc if len(target_cc) == len(source_cc)]
		
		isometries = []
		for target_corner in target_corners:
			try:
				isometries.append(self.find_isometry(other, {source_corner.label: target_corner.label}, respect_fillings))
			except flipper.AssumptionError:
				pass
		
		return isometries
	
	def self_isometries(self):
		''' Returns a list of isometries taking this triangulation to itself. '''
		
		return self.isometries_to(self)
	
	def is_isometric_to(self, other):
		''' Return if there are any orientation preserving isometries from this triangulation to other. '''
		
		assert(isinstance(other, Triangulation))
		
		return len(self.isometries_to(other)) > 0
	
	# Laminations we can build on this triangulation.
	def lamination(self, geometric, algebraic=None, remove_peripheral=True):
		''' Return a new lamination on this surface assigning the specified weight to each edge. '''
		
		if remove_peripheral:
			# Compute how much peripheral component there is on each corner class.
			# This will also check that the triangle inequalities are satisfied. When
			# they fail one of peripheral.values() is negative, which is non-zero and
			# so triggers the correction.
			def dual_weight(corner):
				''' Return double the weight of normal arc corresponding to the given corner. '''
				
				return geometric[corner.indices[1]] + geometric[corner.indices[2]] - geometric[corner.index]
			
			peripheral = dict((vertex, min(dual_weight(corner) for corner in self.corner_class_of_vertex(vertex))) for vertex in self.vertices)
			if any(peripheral.values()):  # Is there any to add / remove?
				# Really should be geometric[i] - sum(peripheral[v]) / 2 but we can't do division in a ring.
				geometric = [2*geometric[i] - sum(peripheral[v] for v in self.vertices_of_edge(i)) for i in range(self.zeta)]
				if all(entry % 2 == 0 for entry in geometric):
					geometric = [entry // 2 for entry in geometric]
		
		if algebraic is not None:
			return flipper.kernel.Lamination(self, geometric, algebraic)
		else:
			lamination = flipper.kernel.Lamination(self, geometric, [0] * self.zeta)
			# If we have a curve we should compute the algebraic intersection numbers.
			# Note that if the curve is not twistable then its algebraic intersection numbers
			# are all zero and so we can just return lamination.
			if lamination.is_curve() and lamination.is_twistable():
				conjugation = lamination.conjugate_short()
				short_lamination = conjugation(lamination)
				triangulation = short_lamination.triangulation
				
				# Grab the indices of the two edges we meet.
				e1, e2 = [edge_index for edge_index in range(short_lamination.zeta) if short_lamination(edge_index) > 0]
				
				a, b, c, d = triangulation.square_about_edge(e1)
				# If the curve is going vertically through the square then ...
				if short_lamination(a) == 1 and short_lamination(c) == 1:
					# swap the labels round so it goes horizontally.
					e1, e2 = e2, e1
					a, b, c, d = triangulation.square_about_edge(e1)
				elif short_lamination(b) == 1 and short_lamination(d) == 1:
					pass
				
				# Currently short_lamination.algebraic == [0, 0, ..., 0].
				# So we need to build a new short_lamination with the correct algebraic intersection numbers.
				geometric = short_lamination.geometric
				algebraic = [1 if i == e1 else -b.sign() if i == b.index else 0 for i in range(self.zeta)]
				
				new_short_lamination = flipper.kernel.Lamination(triangulation, geometric, algebraic)
				return conjugation.inverse()(new_short_lamination)
			else:
				return lamination
	
	def empty_lamination(self):
		''' Return an empty lamination on this surface. '''
		
		return self.lamination([0] * self.zeta, [0] * self.zeta)
	
	def random_curve(self, num_flips):
		''' Return a random curve on this surface. '''
		
		h = self.id_encoding()
		for _ in range(num_flips):
			T = h.target_triangulation
			h = T.encode_flip(choice(T.flippable_edges())) * h
		
		c = h.target_triangulation.key_curves()[0]
		return h.inverse()(c)
	
	def key_curves(self):
		''' Return a list of curves which fill the underlying surface and include a basis for H_1(S).
		
		As these fill, by Alexander's trick a mapping class is the identity
		if and only if it fixes all of them, including orientation. '''
		
		curves = []
		
		for edge_index in self.indices:
			# Build the curve which is the boundary of a regular neighbourhood of this edge.
			endpoints = self.vertices_of_edge(edge_index)
			
			# If the given edge actually forms a loop then we will build two curves, one for
			# each side of the regular neighbourhood.
			if len(set(endpoints)) == 1:
				left = False
				geometric = [[0] * self.zeta, [0] * self.zeta]
				algebraic = [[0] * self.zeta, [0] * self.zeta]
				for corner in self.corner_class_of_vertex(endpoints[0]):  # They are the same so it doesn't matter which we take.
					index = corner.indices[2]
					if index == edge_index:
						left = not left
					else:
						geometric[0 if left else 1][index] += 1
						algebraic[0 if left else 1][index] += 1 if index == corner.labels[2] else -1
				
				curves.append(self.lamination(geometric[0], algebraic[0]))
				curves.append(self.lamination(geometric[1], algebraic[1]))
			else:
				geometric = [0] * self.zeta
				algebraic = [0] * self.zeta
				for vertex in endpoints:
					for corner in self.corner_class_of_vertex(vertex):
						index = corner.indices[2]
						geometric[index] += 1
						algebraic[index] += 1 if index == corner.labels[2] else -1
				
				geometric[edge_index] = 0
				algebraic[edge_index] = 0
				if not self.is_flippable(edge_index):
					# We also need to zero out the curve enclosing this edge.
					boundary_edge = norm(self.nonflippable_boundary(edge_index))
					geometric[boundary_edge] = 0
					algebraic[boundary_edge] = 0
				
				curves.append(self.lamination(geometric, algebraic))
		
		# Now add in paths to make sure we have the homology basis covered.
		for path in self.homology_basis():
			geometric = [0] * self.zeta
			algebraic = [0] * self.zeta
			for step in path:
				geometric[norm(step)] += 1
				algebraic[norm(step)] += +1 if norm(step) == step else -1
			curves.append(self.lamination(geometric, algebraic))
		
		# Filter out any empty laminations that we get.
		return [curve for curve in curves if not curve.is_empty()]
	
	def id_isometry(self):
		''' Return the isometry representing the identity map. '''
		
		return flipper.kernel.Isometry(self, self, dict((i, i) for i in self.labels))
	
	def id_encoding(self):
		''' Return an encoding of the identity map on this triangulation. '''
		
		return self.id_isometry().encode()
	
	def encode_flip(self, edge_label):
		''' Return an encoding of the effect of flipping the given edge.
		
		The given edge must be flippable. '''
		
		assert(self.is_flippable(edge_label))
		
		new_triangulation = self.flip_edge(edge_label)
		
		return flipper.kernel.EdgeFlip(self, new_triangulation, edge_label).encode()
	
	def encode_spiral(self, edge_label, power):
		''' Return an encoding of the effect of spiraling about the given edge.
		
		The given edge must be spiralable. '''
		
		return flipper.kernel.Spiral(self, self, edge_label, power).encode()
	
	def encode_relabel_edges(self, label_map):
		''' Return an encoding of the effect of flipping the given edge. '''
		
		if isinstance(label_map, (list, tuple)):
			label_map = dict(enumerate(label_map))
		else:
			label_map = dict(label_map)
		
		# Build any missing labels.
		for i in self.indices:
			if i in label_map and ~i in label_map:
				pass
			elif i not in label_map and ~i in label_map:
				label_map[i] = ~label_map[~i]
			elif i in label_map and ~i not in label_map:
				label_map[~i] = ~label_map[i]
			else:
				raise flipper.AssumptionError('Missing new label for %d.' % i)
		
		new_triangulation = self.relabel_edges(label_map)
		
		return flipper.kernel.Isometry(self, new_triangulation, label_map).encode()
	
	def encode(self, sequence, _cache=None):
		''' Return the encoding given by sequence.
		
		This consists of EdgeFlips, Isometries and LinearTransformations. Furthermore there are
		several conventions that allow these to be specified by a smaller amount of information.
		 - An integer x represents EdgeFlip(..., edge_label=x)
		 - A dictionary which has i or ~i as a key (for every i) represents a relabelling.
		 - A dictionary which is missing i and ~i (for some i) represents an isometry back to this triangulation.
		 - None represents the identity isometry.
		
		This sequence is read in reverse in order respect composition. For example:
			self.encode([1, {1: ~2}, 2, 3, ~4])
		is the mapping class which: flips edge ~4, then 3, then 2, then relabels
		back to the starting triangulation via the isometry which takes 1 to ~2 and
		then finally flips edge 1. '''
		
		assert(isinstance(sequence, (list, tuple)))
		
		h = None
		for item in reversed(sequence):
			if isinstance(item, flipper.IntegerType):  # Flip.
				if h is None:
					h = self.encode_flip(item)
				else:
					h = h.target_triangulation.encode_flip(item) * h
			elif isinstance(item, dict):  # Isometry.
				if h is None:
					h = self.encode_relabel_edges(item)
				elif all(i in item or ~i in item for i in self.indices):
					h = h.target_triangulation.encode_relabel_edges(item) * h
				else:  # If some edges are missing then we assume that we must be mapping back to this triangulation.
					h = h.target_triangulation.find_isometry(self, item).encode() * h
			elif item is None:  # Identity isometry.
				if h is None:
					h = self.id_encoding()
				else:
					h = h.target_triangulation.id_encoding() * h
			elif isinstance(item, tuple) and len(item) == 2:  # Spiral
				if h is None:
					h = self.encode_spiral(item[0], item[1])
				else:
					h = h.target_triangulation.encode_spiral(item[0], item[1]) * h
			elif isinstance(item, flipper.kernel.Encoding):  # Encoding.
				if h is None:
					h = item
				else:
					h = item * h
			elif isinstance(item, flipper.kernel.Move):  # Move.
				if h is None:
					h = item.encode()
				else:
					h = item.encode() * h
			else:  # Other.
				if h is None:
					h = item.encode()
				else:
					h = item.encode() * h
		
		# Install a cache if we were given one.
		if _cache is not None: h._cache = _cache
		
		return h
	
	def encode_flips_and_close(self, edge_labels, edge_from_label, edge_to_label):
		''' DEPRECIATED: Return an encoding of the effect of flipping the given sequences of edges followed by an isometry.
		
		The isometry used is the one taking edge_from_label to edge_to_label.
		
		This function has been depreciated in favour of self.encode() and is equivalent to:
			self.encode([{edge_from_label: edge_to_label}] + list(reversed(edge_labels)))
		Note that edge_labels needs to be reversed in order to match the order of
		composition used in self.encode(). '''
		
		assert(isinstance(edge_labels, (list, tuple)))
		assert(all(isinstance(label, flipper.IntegerType) for label in edge_labels))
		assert(isinstance(edge_from_label, flipper.IntegerType))
		assert(isinstance(edge_to_label, flipper.IntegerType))
		
		E = self.encode(list(reversed(edge_labels)))
		return E.target_triangulation.find_isometry(self, {edge_from_label: edge_to_label}).encode() * E
	
	def all_flips(self, depth, prefix=None):
		''' Return all flip sequences of at most the given number of flips. '''
		
		assert(isinstance(depth, flipper.IntegerType))
		
		def generator(T, d, flippable):
			''' Return the sequences of at most d flips from this triangulation where only the specified edges are flippable. '''
			for i in flippable:
				yield [i]
				if d > 1:
					T2 = T.flip_edge(i)
					square = [edge.index for edge in T.square_about_edge(i)]
					new_flippable = [j for j in T2.flippable_edges() if (j > i and j in flippable) or j in square]
					for suffix in generator(T.flip_edge(i), d-1, new_flippable):
						yield [i] + suffix
		
		prefix = [] if prefix is None else list(prefix)
		
		flippable = self.flippable_edges()
		T = self
		for item in prefix:
			square = [edge.index for edge in T.square_about_edge(item)]
			T = T.flip_edge(item)
			flippable = [j for j in T.flippable_edges() if (j > item and j in flippable) or j in square]
		
		for sequence in generator(T, depth - len(prefix), flippable):
			yield prefix + sequence
	
	def all_encodings(self, depth, prefix=None):
		''' Return all encodings that can be defined by at most the given number of flips. '''
		
		prefix = [] if prefix is None else list(prefix)
		
		for sequence in self.all_flips(depth, prefix):
			yield self.encode(list(reversed(sequence)) + list(reversed(prefix)))
	
	def all_mapping_classes(self, depth, prefix=None):
		''' Return all mapping classes that can be defined by at most the given number of flips followed by one isometry. '''
		
		assert(isinstance(depth, flipper.IntegerType))
		
		for encoding in self.all_encodings(depth, prefix):
			for isom in encoding.closing_isometries():
				yield isom.encode() * encoding
	
	def components(self):
		''' Return a list of pairs (connected component of self, embedding). '''
		
		remaining_labels = set(self.labels)
		components = []
		while remaining_labels:
			to_process = [remaining_labels.pop()]
			component = set(to_process)
			while to_process:
				label = to_process.pop()
				neighbours = [~label, self.corner_lookup[label].labels[1]]
				for new_label in neighbours:
					if new_label not in component:
						remaining_labels.remove(new_label)
						component.add(new_label)
						to_process.append(new_label)
			
			components.append(list(component))
		
		return components

def create_triangulation(cls, edge_labels, vertex_labels=None, vertex_states=None):
	''' A helper function for pickling. '''
	
	return cls.from_tuple(edge_labels, vertex_labels, vertex_states)





''' A module for the various different errors that can be raised. '''

class AbortError(Exception):
	''' An exception for aborting computations with.
	
	This is thrown by clicking 'cancel' on a progress box. '''
	
	def __init__(self, message=None):
		super(AbortError, self).__init__()
		self.message = message
	def __str__(self):
		return str(self.message)

class ComputationError(Exception):
	''' An exception for when computations fail. '''
	
	def __init__(self, message=None):
		super(ComputationError, self).__init__()
		self.message = message
	def __str__(self):
		return str(self.message)

class AssumptionError(Exception):
	''' An exception for when an assumptions is false. '''
	
	def __init__(self, message=None):
		super(AssumptionError, self).__init__()
		self.message = message
	def __str__(self):
		return str(self.message)

class ApproximationError(Exception):
	''' An exception for when a calculation has been done to insufficient accuracy. '''
	
	def __init__(self, message=None):
		super(ApproximationError, self).__init__()
		self.message = message
	def __str__(self):
		return str(self.message)

class FatalError(Exception):
	''' An exception for when we reach something mathematically impossible.
	
	This indicates that the implementation of the algorithm differs from the theory. '''
	
	def __init__(self, message=None):
		super(FatalError, self).__init__()
		self.message = message
	def __str__(self):
		return str(self.message)





''' A module for interfacing with CyPari. '''

from __future__ import absolute_import

import flipper

import cypari

def lcm(numbers):
	''' Return the lcm of a list of numbers. '''
	
	x = cypari.gen.pari(1)
	for number in numbers:
		x = cypari.all.pari_gen.lcm(x, cypari.gen.pari(number))
	
	return int(x)

def directed_eigenvector(action_matrix, condition_matrix):
	''' An implementation of flipper.kernel.symboliccomputation.directed_eigenvector() using CyPari.
	
	See the docstring of the above function for further details. '''
	
	x = cypari.gen.pari('x')
	
	M = cypari.gen.pari.matrix(action_matrix.width, action_matrix.height, action_matrix.flatten())
	
	for polynomial in M.charpoly().factor()[0]:
		degree = int(polynomial.poldegree())
		if degree > 1:
			flipper_polynomial_roots = flipper.kernel.Polynomial([int(polynomial.polcoeff(i)) for i in range(degree+1)]).real_roots()
			
			if len(flipper_polynomial_roots) > 0:
				# We need only consider the largest root as it has to be >=1 and bigger than all of its Galois conjugates.
				flipper_polynomial_root = max(flipper_polynomial_roots)
				
				if flipper_polynomial_root >= 1:
					# Compute the kernel:
					a = x.Mod(polynomial)
					kernel_basis = (M - a).matker()
					
					basis = [[[entry.lift().polcoeff(i) for i in range(degree)] for entry in v] for v in kernel_basis]
					scale = lcm([int(coeff.denominator()) for v in basis for entry in v for coeff in entry])
					scaled_basis = [[[int(int(coeff.numerator()) * scale) / int(coeff.denominator()) for coeff in entry] for entry in v] for v in basis]
					
					N = flipper.kernel.NumberField(flipper_polynomial_root)
					flipper_basis_matrix = flipper.kernel.Matrix([[N.element(entry) for entry in v] for v in scaled_basis])
					
					if len(flipper_basis_matrix) == 1:  # If rank(kernel) == 1.
						[flipper_eigenvector] = flipper_basis_matrix
						if flipper.kernel.matrix.nonnegative(flipper_eigenvector) and condition_matrix.nonnegative_image(flipper_eigenvector):
							return N.lmbda, flipper_eigenvector
					else:
						T = (flipper.kernel.id_matrix(condition_matrix.width).join(condition_matrix)) * flipper_basis_matrix.transpose()
						try:
							linear_combination = T.find_vector_with_nonnegative_image()
							return N.lmbda, flipper_basis_matrix.transpose()(linear_combination)
						except flipper.AssumptionError:  # Eigenspace is disjoint from the cone.
							pass
	
	raise flipper.ComputationError('No interesting eigenvalues in cell.')





from __future__ import print_function

import flipper

NT_TYPE_PERIODIC = flipper.kernel.encoding.NT_TYPE_PERIODIC
NT_TYPE_REDUCIBLE = flipper.kernel.encoding.NT_TYPE_REDUCIBLE
NT_TYPE_PSEUDO_ANOSOV = flipper.kernel.encoding.NT_TYPE_PSEUDO_ANOSOV

def main(verbose=False):
	if verbose: print('Running encoding tests.')
	
	S = flipper.load('S_1_1')
	f, g = S.mapping_class('aB'), S.mapping_class('bA')  # Some pseudo-Anosov ones.
	h, i = S.mapping_class('ab'), S.mapping_class('')  # Some finite order ones.
	
	tests = [
		h != i,
		h**6 == i,
		h**-3 == h**3,
		h**0 == i,
		h.order() == 6,
		f != g,
		f.is_conjugate_to(g),
		not f.is_conjugate_to(h),
		not h.is_conjugate_to(i),
		]
	
	if not all(tests):
		if verbose: print(tests)
		return False
	
	# Add more examples here.
	examples = [
		('S_1_1', 'a', NT_TYPE_REDUCIBLE),
		('S_1_1', 'ab', NT_TYPE_PERIODIC),
		('S_1_1', 'aB', NT_TYPE_PSEUDO_ANOSOV),
		('S_1_2', 'a', NT_TYPE_REDUCIBLE),
		('S_1_2', 'b', NT_TYPE_REDUCIBLE),
		('S_1_2', 'c', NT_TYPE_REDUCIBLE),
		('S_1_2', 'aB', NT_TYPE_REDUCIBLE),
		('S_1_2', 'bbaCBAaBabcABB', NT_TYPE_REDUCIBLE),
		('S_1_2', 'aCBACBacbaccbAaAcAaBBcCcBBcCaBaaaABBabBcaBbCBCbaaa', NT_TYPE_PSEUDO_ANOSOV),
		('S_2_1', 'aaabcd', NT_TYPE_PSEUDO_ANOSOV),
		('S_2_1', 'abcdeF', NT_TYPE_PSEUDO_ANOSOV),
		#('E_12', 'aaaaBBp', NT_TYPE_PSEUDO_ANOSOV),
		#('E_12', 'aaBaaBBp', NT_TYPE_REDUCIBLE),
		#('E_12', 'aaaaBBaBaBp', NT_TYPE_PSEUDO_ANOSOV),
		]
	
	for surface, word, mapping_class_type in examples:
		if verbose: print(word)
		S = flipper.load(surface)
		h = S.mapping_class(word)
		# Check that these mapping classes have the correct Nielsen--Thurston type.
		if h.nielsen_thurston_type() != mapping_class_type:
			return False
		# Check that packaging and unpackaging works.
		if h.package() != h.source_triangulation.encode(h.package()).package():
			return False
	
	return True

if __name__ == '__main__':
	print(main(verbose=True))





try:
	import Tkinter as TK
except ImportError:  # Python 3.
	try:
		import tkinter as TK
	except ImportError:
		raise ImportError('Tkinter not available.')
from time import sleep

textFont1 = ('Arial', 10, 'bold italic')
textFont2 = ('Arial', 16, 'bold')
textFont3 = ('Arial', 8, 'bold')

class LabelWidget(TK.Entry):
	def __init__(self, master, x, y, text, **options):
		self.text = TK.StringVar(value=text)
		self.text.set(text)
		#super(LabelWidget, self).__init__(self, master=master)
		if 'state' not in options: options['state'] = 'readonly'
		options['textvariable'] = self.text
		TK.Entry.__init__(self, master=master, **options)
		self.grid(column=x, row=y)

class Grid(TK.Frame):
	def __init__(self, parent, data, **options):
		TK.Frame.__init__(self, parent, **options)
		
		assert(len(data) > 0)
		assert(all(len(row) == len(data[0]) for row in data))
		
		self.parent = parent
		self._data = data
		
		self._num_rows = len(self._data)
		self._num_columns = len(self._data[0])
		
		self._variables = [[TK.StringVar(value=entry) for entry in row] for row in self._data]
		self._entries = [[TK.Entry(self, textvariable=var) for var in row] for row in self._variables]
		self._entries = [[LabelWidget(self, i, j, text=entry) for j, entry in enumerate(row)] for i, row in enumerate(self._data)]
		for i in range(self._num_rows):
			for j in range(self._num_columns):
				self._entries[i][j].grid(column=j, row=i, sticky='nsew')
		
		for i in range(self._num_rows):
			self.grid_rowconfigure(i, weight=1)
		for j in range(self._num_columns):
			self.grid_columnconfigure(j, weight=1)

class TestApp(object):
	def __init__(self, parent):
		
		self.parent = parent
		data = [['A', 'B', 'C', 'D'] for i in range(10)]
		self.grid = Grid(self.parent, data)
		self.grid.pack(padx=6, pady=6, fill='both', expand=True)







class EntryGrid(TK.Tk):
	''' Dialog box with Entry widgets arranged in columns and rows.'''
	def __init__(self, colList, rowList, title='Entry Grid'):
		self.cols = colList[:]
		self.colList = colList[:]
		self.colList.insert(0, '')
		self.rowList = rowList
		TK.Tk.__init__(self)
		self.title(title)
	
		self.mainFrame = TK.Frame(self)
		self.mainFrame.config(padx='3.0m', pady='3.0m')
		self.mainFrame.grid()
		self.make_header()
	
		self.gridDict = {}
		for i in range(1, len(self.colList)):
			for j in range(len(self.rowList)):
				w = EntryWidget(self.mainFrame, i, j+1)
				self.gridDict[(i-1,j)] = w.value
				def handler(event, col=i-1, row=j):
					return self.__entryhandler(col, row)
				w.bind(sequence='<FocusOut>', func=handler)
		self.mainloop()
	
	def make_header(self):
		self.hdrDict = {}
		for i, label in enumerate(self.colList):
			def handler(event, col=i, row=0, text=label):
				return self.__headerhandler(col, row, text)
			w = LabelWidget(self.mainFrame, i, 0, label)
			self.hdrDict[(i,0)] = w
			w.bind(sequence='<KeyRelease>', func=handler)
	
		for i, label in enumerate(self.rowList):
			def handler(event, col=0, row=i+1, text=label):
				return self.__headerhandler(col, row, text)
			w = LabelWidget(self.mainFrame, 0, i+1, label)
			self.hdrDict[(0,i+1)] = w
			w.bind(sequence='<KeyRelease>', func=handler)
	
	def __entryhandler(self, col, row):
		s = self.gridDict[(col,row)].get()
		if s.upper().strip() == 'EXIT':
			self.destroy()
		elif s.strip():
			print(s)
	
	def __headerhandler(self, col, row, text):
		''' has no effect when Entry state=readonly '''
		self.hdrDict[(col,row)].text.set(text)
	
	def get(self, x, y):
		return self.gridDict[(x,y)].get()
	
	def set(self, x, y, v):
		self.gridDict[(x,y)].set(v)
		return v

if __name__ == '__main__':
	root = TK.Tk()
	app = TestApp(root)
	root.mainloop()





import flipper
from random import seed

def main(verbose=False):
	seed(1)  # To make this test deterministic.
	
	S = flipper.load('S_1_2')
	
	all_words = set(S.all_words(5, equivalence='none'))
	# There should be 8^5 + 8^4 + ... + 8^0 = 37449 words.
	if verbose: print('Constructed %d words.' % len(all_words))
	for _ in range(10):
		word = S.random_word(5)
		if word not in all_words:
			return False
	
	if not S.mapping_class('abababababab') == S.mapping_class('xx'):
		return False
	
	return True

if __name__ == '__main__':
	print(main(verbose=True))




''' A module for representing a splitting sequence of a lamination.

Provides one class: SplittingSequence. '''

import flipper

class SplittingSequence(object):
	''' This represents a sequence of flips of an Triangulation. '''
	def __init__(self, preperiodic, mapping_class, dilatation, lamination):
		assert(isinstance(preperiodic, flipper.kernel.Encoding))
		assert(isinstance(mapping_class, flipper.kernel.Encoding))
		# assert(isinstance(dilatation, flipper.kernel.NumberFieldElement))
		assert(isinstance(lamination, flipper.kernel.Lamination))
		
		# We wont bother to save encoding as it can always be reconstructed by:
		#  self.mapping_class.inverse() * self.preperiodic
		
		self.preperiodic = preperiodic
		self.mapping_class = mapping_class
		self.dilatation = dilatation
		self.lamination = lamination
		
		self.triangulation = self.lamination.triangulation
		
		# Write some things into the cache.
		# Hmmm, this assumes we're not taking roots.
		self.mapping_class._cache['invariant_lamination'] = (self.dilatation, self.lamination)

class SplittingSequences(object):
	''' This represents a sequence of flips of an Triangulation. '''
	def __init__(self, encoding, isometries, index, dilatation, lamination):
		assert(isinstance(encoding, flipper.kernel.Encoding))
		
		assert(isinstance(index, flipper.IntegerType))
		# assert(isinstance(dilatation, flipper.kernel.NumberFieldElement))
		assert(isinstance(lamination, flipper.kernel.Lamination))
		
		# We wont bother to save encoding as it can always be reconstructed by:
		#  self.mapping_class.inverse() * self.preperiodic
		
		self.encoding = encoding
		self.isometries = isometries
		self.index = index
		self.dilatation = dilatation
		self.lamination = lamination
		
		self.preperiodic = self.encoding[-self.index:]
		self.open_periodic = self.encoding[:-self.index]
	
	def __iter__(self):
		for isometry in self.isometries:
			# We will reverse the direction of self.mapping_class so that self.lamination is the stable lamination.
			yield SplittingSequence(self.preperiodic, (isometry.encode() * self.open_periodic).inverse(), self.dilatation, self.lamination)
		
		return






''' A module for the standard types.

Provides a standard reference for the type of integers, strings, etc. as these depend on the
version of Python being used. This is similar to the six module. '''

import sys
import io

if sys.version_info >= (3, 0):
	StringType = str
	IntegerType = (int,)
	FileType = io.TextIOWrapper
else:
	# In Python 2.x an integer can be an int or long (Python automatically switches to longs when required).
	StringType = str
	IntegerType = (int, long)
	FileType = file





from __future__ import print_function

import flipper

import importlib

def main(verbose=False):
	''' Runs all of the tests in the tests directory. '''
	
	failed_tests = []
	for test_name in dir(flipper.test):
		if not test_name.startswith('_'):
			test = importlib.import_module('flipper.test.%s' % test_name)
			print('Running %s test...' % test_name)
			result = test.main(verbose=verbose)
			print('\tPassed' if result else '\tFAILED')
			if not result:
				failed_tests.append(test_name)
	
	print('Finished testing.')
	if len(failed_tests) > 0:
		print('\tFAILED TESTS:')
		for test_name in failed_tests:
			print('\t%s' % test_name)
	else:
		print('\tAll tests passed.')

if __name__ == '__main__':
	main()





''' A module for storing constants that should only ever be computed once.

We need to put these in a special place so that flipper can compute these after everything
else has been initialised. '''

import flipper

# Create special permutations.
Permutation = flipper.kernel.Permutation
flipper.kernel.permutation.PERM3 = flipper.kernel.permutation.all_permutations(3)
flipper.kernel.permutation.PERM3_INVERSE = {perm: perm.inverse() for perm in flipper.kernel.permutation.PERM3}
flipper.kernel.permutation.PERM3_LOOKUP = dict((perm, index) for index, perm in enumerate(flipper.kernel.permutation.PERM3))
flipper.kernel.permutation.TRANSITION_PERM3_LOOKUP = {
	(0, 0): Permutation([0, 2, 1]),
	(0, 1): Permutation([1, 0, 2]),
	(0, 2): Permutation([2, 1, 0]),
	(1, 0): Permutation([1, 0, 2]),
	(1, 1): Permutation([2, 1, 0]),
	(1, 2): Permutation([0, 2, 1]),
	(2, 0): Permutation([2, 1, 0]),
	(2, 1): Permutation([0, 2, 1]),
	(2, 2): Permutation([1, 0, 2])
}





''' A module for representing approximations of real algebraic numbers.

Provides one class: AlgebraicApproximation. '''

# Suppose that f(x) = a_n x^n + ... + a_0 \in ZZ[x] is a (not necessarily irreducible) polynomial with a_n != 0. We define
# h(f) := log(max(|a_n|)) to be its height and deg(f) := n to be its degree.
#
# Let K be a number field and x_0 \in K be an algebraic number. We define h(x_0) := h(minpoly(x_0)) to be its the height
# and deg(x_0) := deg(minpoly(x_0)) to be its degree.
#
# We use the following facts:
#	1a) For x_0, x_1 \in K, h(x_0 +/- x_1) <= h(x_0) + h(x_1) + 1,
#	 b) h(x_0 * x_1) <= h(x_0) + h(x_1) and
#	 c) h(1 / x_0) == h(x_0) [Waldschmidt "Diophantine approximation on linear algebraic groups", Property 3.3].
#	2) If 0 != x_0 \in K is a root of f(x) = a_n x^n + ... + a_0 then
#		|x_0| >= 1 / sum(|a_i / a_0|) [Basu et al. "Algorithms in Real Algebraic Geometry", Lemma 10.3].
#
# An immediate consequence of 1) is that if x_0 \in K is a root of f \in ZZ[x] then h(x_0) <= h(f) + 2 \deg(f).
#
# From 2) it follows that so long as the accuracy of the interval of an AlgebraicApproximation is at least
#	-log(1 / sum(|a_i / a_0|)) = log(sum(|a_i / a_0|)) <= log(sum(|a_i|)) <= log(deg(x_0) * H(f)) <= log(deg(x_0)) + h(x_0)
# then it uniquely determines an algebraic number.
#
# Thus by knowing a sufficiently accurate approximation of x_0 we can determine if x_0 > 0. Combining this with 1) we can
# therefore determine if x_0 > x_1 by determining if (x_0 - x_1) > 0.

import flipper

from math import log10 as log
LOG_2 = log(2)

class AlgebraicApproximation(object):
	''' This represents an algebraic number.
	
	It is specified by an interval which contains the number, an upper
	bound on the log of its degree and its height.
	
	If the interval is not sufficiently small, in terms of the log_degree
	and height bounds, to uniquely determine the algebraic number then
	an ApproximationError will be raised.
	
	>>> I = AlgebraicApproximation(flipper.kernel.Interval(141421356, 141421357, 8), 2, 2)
	>>> I
	1.414213?
	>>> J = AlgebraicApproximation.from_tuple('1.41421356', 2, 2)
	>>> I == J
	True
	>>> AlgebraicApproximation.from_tuple('1.41', 2, 2)  # doctest: +ELLIPSIS
	Traceback (most recent call last):
	    ...
	ApproximationError: ...
	'''
	def __init__(self, interval, log_degree, height):
		assert(isinstance(interval, flipper.kernel.Interval))
		
		self.interval = interval
		self.log_degree = log_degree
		self.accuracy = self.interval.accuracy
		self.height = height
		self.accuracy_needed = self.log_degree + self.height
		# An algebraic approximation is good if it is known to more places
		# than its accuracy needed. That is if self.interval.accuracy >= self.accuracy_needed.
		if self.accuracy < self.accuracy_needed:
			raise flipper.kernel.ApproximationError('An algebraic number with log(degree) at most %0.3f and height at most %0.3f requires an interval with accuracy at least %0.3f, not %d.' % (self.log_degree, self.height, self.accuracy_needed, self.accuracy))
	
	@classmethod
	def from_tuple(cls, string, degree, height):
		''' A short way of constructing AlgebraicApproximations from a string and degree and height bounds. '''
		
		return cls(flipper.kernel.Interval.from_string(string), log(max(degree, 1)), height)
	
	@classmethod
	def from_int(cls, integer, accuracy=None):
		''' A short way of constructing AlgebraicApproximations from a string and degree and height bounds. '''
		
		return cls(flipper.kernel.Interval.from_int(integer, accuracy), 1, flipper.kernel.height_int(integer))
	
	def __repr__(self):
		return str(self)
	def __str__(self):
		return self.approximate_string(6)
	def approximate_string(self, accuracy=None):
		''' Return a string approximating this algebraic number. '''
		
		return self.interval.approximate_string(accuracy)
	
	def __float__(self):
		return float(self.interval)
	def __int__(self):
		# This should return the largest integer less than or equal to self.
		x = int(self.interval)  # This gets pretty close.
		# However self.interval.lower is slightly smaller than self. Hence if self is an integer
		# it's possible that x == int(self) - 1 so we will just check if x+1 also works.
		return x+1 if x+1 <= self else x
	
	def change_denominator(self, new_denominator):
		''' Return a new approximation of this algebraic number with the given denominator. '''
		
		assert(isinstance(new_denominator, flipper.IntegerType))
		
		return AlgebraicApproximation(self.interval.change_denominator(new_denominator), self.log_degree, self.height)
	
	def change_accuracy(self, new_accuracy):
		''' Return a new approximation of this algebraic number with the given accuracy.
		
		The new_accuracy must be at most self.accuracy.
		
		>>> I.change_accuracy(5)
		1.414?
		'''
		
		assert(isinstance(new_accuracy, flipper.IntegerType))
		
		return AlgebraicApproximation(self.interval.change_accuracy(new_accuracy), self.log_degree, self.height)
	
	def simplify(self):
		''' Return a new approximation of this algebraic number with the given accuracy.
		
		>>> I.simplify()
		1.414213?
		'''
		
		return AlgebraicApproximation(self.interval.simplify(), self.log_degree, self.height)
	
	def __neg__(self):
		return AlgebraicApproximation(-self.interval, self.log_degree, self.height)
	def __abs__(self):
		return self if self >= 0 else -self
	def __add__(self, other):
		if isinstance(other, AlgebraicApproximation):
			return AlgebraicApproximation(self.interval + other.interval, self.log_degree + other.log_degree, self.height + other.height + LOG_2)
		elif isinstance(other, flipper.IntegerType):
			return AlgebraicApproximation(self.interval + other, self.log_degree, self.height + flipper.kernel.height_int(other) + LOG_2)
		else:
			return NotImplemented
	def __radd__(self, other):
		return self + other
	def __sub__(self, other):
		if isinstance(other, AlgebraicApproximation):
			return AlgebraicApproximation(self.interval - other.interval, self.log_degree + other.log_degree, self.height + other.height + LOG_2)
		elif isinstance(other, flipper.IntegerType):
			return AlgebraicApproximation(self.interval - other, self.log_degree, self.height + flipper.kernel.height_int(other) + LOG_2)
		else:
			return NotImplemented
	def __rsub__(self, other):
		return -(self - other)
	def __mul__(self, other):
		if isinstance(other, AlgebraicApproximation):
			return AlgebraicApproximation(self.interval * other.interval, self.log_degree + other.log_degree, self.height + other.height)
		elif isinstance(other, flipper.IntegerType):
			return AlgebraicApproximation(self.interval * other, self.log_degree, self.height + flipper.kernel.height_int(other))
		else:
			return NotImplemented
	def __rmul__(self, other):
		return self * other
	def __pow__(self, power):
		if power == 0:
			return AlgebraicApproximation(flipper.kernel.Interval(1, 1, 0), 0, 1)
		if power > 0:
			sqrt = self**(power//2)
			square = sqrt * sqrt
			if power % 2 == 1:
				return self * square
			else:
				return square
	def __div__(self, other):
		return self.__truediv__(other)
	def __truediv__(self, other):
		if isinstance(other, AlgebraicApproximation):
			return AlgebraicApproximation(self.interval / other.interval, self.log_degree + other.log_degree, self.height + other.height)
		elif isinstance(other, flipper.IntegerType):
			return AlgebraicApproximation(self.interval / other, self.log_degree, self.height + flipper.kernel.height_int(other))
		else:
			return NotImplemented
	def __rdiv__(self, other):
		if isinstance(other, flipper.IntegerType):
			return AlgebraicApproximation(other / self.interval, self.log_degree, self.height + flipper.kernel.height_int(other))
		else:
			return NotImplemented
	def __rtruediv__(self, other):
		return self.__rdiv__(other)
	
	def sign(self):
		''' Return the sign of the underlying algebraic number.
		
		>>> I.sign()
		1
		'''
		
		if self.interval.lower > 0:
			return +1
		elif self.interval.upper < 0:
			return -1
		else:
			return 0
	
	def __lt__(self, other):
		return (self - other).sign() < 0
	def __eq__(self, other):
		return (self - other).sign() == 0
	def __gt__(self, other):
		return (self - other).sign() > 0
	def __le__(self, other):
		return not (self > other)
	def __ne__(self, other):
		return not (self == other)
	def __ge__(self, other):
		return not (self < other)

def doctest_globs():
	''' Return the globals needed to run doctest on this module. '''
	
	I = flipper.kernel.AlgebraicApproximation(flipper.kernel.Interval(141421356, 141421357, 8), 2, 2)
	J = flipper.kernel.AlgebraicApproximation.from_tuple('1.41421356', 2, 2)
	
	return {'I': I, 'J': J}





from __future__ import print_function

import flipper

def main(verbose=False):
	if verbose: print('Running triangulation tests.')
	
	# These need to be changed if the standard example triangulations ever change.
	num_isometries = [
		('S_0_4', 2),
		('S_1_1', 6),
		('S_1_2', 4),
		('S_2_1', 2),
		('S_3_1', 2),
		('E_12', 12),
		('E_24', 24),
		('E_36', 36)
		]
	
	# Check that every triangulation has the correct number of isometries to itself.
	for surface, num_isoms in num_isometries:
		if verbose: print('Checking: %s' % surface)
		S = flipper.load(surface)
		T = S.triangulation
		if len(T.self_isometries()) != num_isoms > 0:
			return False
		# Check that isomorphism signatures work.
		T2 = flipper.triangulation_from_iso_sig(T.iso_sig())
		if not T.is_isometric_to(T2):
			return False
		if T.iso_sig() != T2.iso_sig():
			return False
	
	return True

if __name__ == '__main__':
	print(main(verbose=True))





''' A module for representing triangulations of 3--manifolds.

Provides two classes: Tetrahedron and Triangulation3. '''

# We use right-handed tetrahedra, see SnapPy/headers/kernel_typedefs.h.
#
# We follow the orientation conventions in SnapPy/headers/kernel_typedefs.h L:154
# and SnapPy/kernel/peripheral_curves.c.

import flipper

from itertools import combinations, product

# Edge veerings:
VEERING_UNKNOWN, VEERING_LEFT, VEERING_RIGHT = 'Unknown', 'Left', 'Right'
# Peripheral curve types:
LONGITUDES, MERIDIANS, TEMPS = 0, 1, 2  # These are used for indexing so need to be integers.
PERIPHERAL_TYPES = [LONGITUDES, MERIDIANS, TEMPS]
# Tetrahedron geometry:
# This order was chosen so they appear ordered anti-clockwise from the cusp.
VERTICES_MEETING = {0: (1, 2, 3), 1: (0, 3, 2), 2: (0, 1, 3), 3: (0, 2, 1)}
# If you enter cusp a through side b then EXIT_CUSP_LEFT[(a, b)] is the side to your left.
EXIT_CUSP_LEFT = {
	(0, 1): 3, (0, 2): 1, (0, 3): 2,
	(1, 0): 2, (1, 2): 3, (1, 3): 0,
	(2, 0): 3, (2, 1): 0, (2, 3): 1,
	(3, 0): 1, (3, 1): 2, (3, 2): 0
	}
# Similarly EXIT_CUSP_RIGHT[(a, b)] is the side to your right.
EXIT_CUSP_RIGHT = {
	(0, 1): 2, (0, 2): 3, (0, 3): 1,
	(1, 0): 3, (1, 2): 0, (1, 3): 2,
	(2, 0): 1, (2, 1): 3, (2, 3): 0,
	(3, 0): 2, (3, 1): 0, (3, 2): 1
	}

class Tetrahedron(object):
	''' This represents a tetrahedron. '''
	def __init__(self, label):
		assert(isinstance(label, flipper.IntegerType))
		
		self.label = label
		self.glued_to = [None] * 4  # Each entry is either: None or (Tetrahedron, permutation).
		self.cusp_indices = [-1, -1, -1, -1]
		self.peripheral_curves = [[[0, 0, 0, 0] for _ in range(4)] for _ in PERIPHERAL_TYPES]
		self.edge_labels = dict((vertex_pair, VEERING_UNKNOWN) for vertex_pair in combinations(range(4), r=2))
		self.vertex_labels = [None, None, None, None]
	
	def __repr__(self):
		return str(self)
	def __str__(self):
		return '%s --> %s' % (self.label, ', '.join('%d: %s' % (x[0].label, x[1]) if x is not None else 'X' for x in self.glued_to))
	
	def glue(self, side, target, permutation):
		''' Glue the given side of this tetrahedron to target via the given permutation.
		
		You are not supposed to unglue tetrahedra as they automatically install a lot
		of additional structure.'''
		
		if self.glued_to[side] is None:
			assert(self.glued_to[side] is None)
			assert(target.glued_to[permutation(side)] is None)
			assert(not permutation.is_even())
			
			self.glued_to[side] = (target, permutation)
			target.glued_to[permutation(side)] = (self, permutation.inverse())
			
			# Move across the edge veerings too, is possible.
			for a, b in combinations(VERTICES_MEETING[side], 2):
				x, y = permutation(a), permutation(b)
				my_edge_veering = self.get_edge_label(a, b)
				his_edge_veering = target.get_edge_label(x, y)
				if my_edge_veering == VEERING_UNKNOWN and his_edge_veering != VEERING_UNKNOWN:
					self.set_edge_label(a, b, his_edge_veering)
				elif my_edge_veering != VEERING_UNKNOWN and his_edge_veering == VEERING_UNKNOWN:
					target.set_edge_label(x, y, my_edge_veering)
		else:
			assert((target, permutation) == self.glued_to[side])
	
	def get_edge_label(self, a, b):
		''' Return the label on edge (a) -- (b) of this tetrahedron. '''
		
		a, b = min(a, b), max(a, b)
		return self.edge_labels[(a, b)]
	
	def set_edge_label(self, a, b, value):
		''' Set the label on edge (a) -- (b) of this tetrahedron. '''
		
		a, b = min(a, b), max(a, b)
		self.edge_labels[(a, b)] = value
	
	def snappy_string(self):
		''' Return the SnapPy string describing this tetrahedron. '''
		
		strn = ''
		strn += '%4d %4d %4d %4d \n' % tuple([tetrahedra.label for tetrahedra, gluing in self.glued_to])
		strn += ' %4s %4s %4s %4s\n' % tuple([gluing.compressed_string() for tetrahedra, gluing in self.glued_to])
		strn += '%4d %4d %4d %4d \n' % tuple(self.cusp_indices)
		strn += ' %2d %2d %2d %2d %2d %2d %2d %2d %2d %2d %2d %2d %2d %2d %2d %2d\n' % tuple(cusp for meridian in self.peripheral_curves[MERIDIANS] for cusp in meridian)
		strn += '  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0\n'
		strn += ' %2d %2d %2d %2d %2d %2d %2d %2d %2d %2d %2d %2d %2d %2d %2d %2d\n' % tuple(cusp for longitude in self.peripheral_curves[LONGITUDES] for cusp in longitude)
		strn += '  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0\n'
		strn += '  0.000000000000 0.000000000000\n'
		return strn
	
	def __snappy__(self):
		return self.snappy_string()

class Triangulation3(object):
	''' This represents triangulation, that is a collection of tetrahedra. '''
	def __init__(self, num_tetrahedra):
		assert(isinstance(num_tetrahedra, flipper.IntegerType))
		
		self.num_tetrahedra = num_tetrahedra
		self.tetrahedra = [Tetrahedron(i) for i in range(self.num_tetrahedra)]
		self.cusps = None
		self.num_cusps = -1
	
	def __repr__(self):
		return str(self)
	def __str__(self):
		return '\n'.join(str(tetrahedron) for tetrahedron in self)
	
	def __iter__(self):
		return iter(self.tetrahedra)
	
	def clear_temp_peripheral_structure(self):
		''' Remove all TEMP peripheral curves. '''
		
		for tetrahedron in self:
			tetrahedron.peripheral_curves[TEMPS] = [[0, 0, 0, 0] for _ in range(4)]
	
	def is_closed(self):
		''' Return if this triangulation is closed. '''
		
		return all(tetrahedron.glued_to[side] is not None for tetrahedron in self for side in range(4))
	
	def is_veering(self):
		''' Return if this triangulation is veering. '''
		
		# Hmmm, this only checks the local condition. This test appears to be
		# assuming that there are no edges labelled VEERING_UNKNOWN.
		for tetrahedron in self:
			for side in range(4):
				if tetrahedron.glued_to[side] is not None:
					target, permutation = tetrahedron.glued_to[side]
					for a, b in combinations(VERTICES_MEETING[side], 2):
						if tetrahedron.get_edge_label(a, b) != target.get_edge_label(permutation(a), permutation(b)):
							return False
		
		return True
	
	def cusp_identification_map(self):
		''' Return a dictionary pairing the sides of the peripheral triangles. '''
		
		cusp_pairing = dict()
		for tetrahedron in self.tetrahedra:
			for side in range(4):
				for other in VERTICES_MEETING[side]:
					neighbour_tetrahedron, permutation = tetrahedron.glued_to[other]
					neighbour_side, neighbour_other = permutation(side), permutation(other)
					cusp_pairing[(tetrahedron, side, other)] = (neighbour_tetrahedron, neighbour_side, neighbour_other)
		
		return cusp_pairing
	
	def assign_cusp_indices(self):
		''' Assign the tetrahedra in this triangulation their cusp indices and return the list of corners in the same cusp.
		
		This triangulation must be closed. '''
		
		assert(self.is_closed())
		
		cusp_pairing = self.cusp_identification_map()
		
		self.cusps = []
		remaining_vertices = list(product(self, range(4)))
		while remaining_vertices:
			# We do a depth first search to find all vertices in this cusp.
			new_vertex = remaining_vertices.pop()
			new_vertex_class = [new_vertex]
			# This is a stack of triangles that may still have consequences.
			to_explore = [new_vertex]
			while to_explore:
				current_tetrahedron, current_side = to_explore.pop()
				for side in VERTICES_MEETING[current_side]:
					neighbour_tetrahedron, neighbour_side, _ = cusp_pairing[(current_tetrahedron, current_side, side)]
					neighbour_vertex = neighbour_tetrahedron, neighbour_side
					if neighbour_vertex in remaining_vertices:
						to_explore.append(neighbour_vertex)
						new_vertex_class.append(neighbour_vertex)
						remaining_vertices.remove(neighbour_vertex)
			
			self.cusps.append(new_vertex_class)
		
		# Then iterate through each one assigning cusp indices.
		for index, vertex_class in enumerate(self.cusps):
			for tetrahedron, side in vertex_class:
				tetrahedron.cusp_indices[side] = index
		
		self.num_cusps = len(self.cusps)
		
		return self.cusps
	
	def intersection_number(self, peripheral_type_a, peripheral_type_b, cusp=None):
		''' Return the algebraic intersection number between the specified peripheral types.
		
		See SnapPy/kernel_code/intersection_numbers.c for more information.
		
		Convention:
		    B
		    ^
		    |
		 ---+---> A
		    |
		    |
		has intersection <A, B> := +1. '''
		
		# This is the number of strands flowing from A to B. It is negative if they go in the opposite direction.
		flow = lambda A, B: 0 if (A < 0) == (B < 0) else (A if (A < 0) != (A < -B) else -B)
		
		if cusp is None: cusp = list(product(self.tetrahedra, range(4)))
		
		intersection_number = 0
		for tetrahedron, side in cusp:
			periph_a = tetrahedron.peripheral_curves[peripheral_type_a]
			periph_b = tetrahedron.peripheral_curves[peripheral_type_b]
			
			# Count intersection numbers along edges.
			for other in VERTICES_MEETING[side]:
				# Only count crossings where periph_a is entering to avoid double counting.
				if periph_a[side][other] > 0:
					intersection_number -= periph_a[side][other] * periph_b[side][other]
			
			# and then within each face.
			for other in VERTICES_MEETING[side]:
				left = EXIT_CUSP_LEFT[(side, other)]
				right = EXIT_CUSP_RIGHT[(side, other)]
				
				# Even though this looks weird it is the correct pattern.
				intersection_number += flow(periph_a[side][other], periph_a[side][left]) * flow(periph_b[side][other], periph_b[side][left])
				intersection_number += flow(periph_a[side][other], periph_a[side][right]) * flow(periph_b[side][other], periph_b[side][left])
		
		return intersection_number
	
	def install_peripheral_curves(self):
		''' Assign a longitude and meridian to each cusp.
		
		Assumes (and checks) the link of each vertex is a torus. '''
		
		# Install the cusp indices.
		self.assign_cusp_indices()
		cusp_pairing = self.cusp_identification_map()
		
		# Blank out the longitudes and meridians.
		for peripheral_type in [LONGITUDES, MERIDIANS]:
			for tetrahedron in self.tetrahedra:
				for side in range(4):
					for other in range(4):
						tetrahedron.peripheral_curves[peripheral_type][side][other] = 0
		
		# Install a longitude and meridian on each cusp one at a time.
		for cusp in self.cusps:
			# Build a triangulation of the cusp neighbourhood.
			label = 0
			edge_label_map = dict()
			for tetrahedron, side in cusp:
				for other in VERTICES_MEETING[side]:
					key = (tetrahedron, side, other)
					if key not in edge_label_map:
						edge_label_map[key] = label
						edge_label_map[cusp_pairing[key]] = ~label
						label += 1
			
			edge_labels = [[edge_label_map[(tetrahedron, side, other)] for other in VERTICES_MEETING[side]] for tetrahedron, side in cusp]
			T = flipper.kernel.create_triangulation(edge_labels)
			
			if T.genus != 1:
				raise flipper.AssumptionError('Non torus vertex link.')
			
			# Get a basis for H_1.
			homology_basis_paths = T.homology_basis()
			
			# Install the longitude and meridian. # !?! Double check and optimise this.
			for peripheral_type in [LONGITUDES, MERIDIANS]:
				last = homology_basis_paths[peripheral_type][-1]
				# Find a starting point.
				for tetrahedron, side in cusp:
					for x in VERTICES_MEETING[side]:
						if edge_label_map[(tetrahedron, side, x)] == last:
							current_tetrahedron, current_side, arrive = tetrahedron, side, x
							break
				
				for other in homology_basis_paths[peripheral_type]:
					for a in VERTICES_MEETING[current_side]:
						if edge_label_map[(current_tetrahedron, current_side, a)] == ~other:
							leave = a
							current_tetrahedron.peripheral_curves[peripheral_type][current_side][arrive] += 1
							current_tetrahedron.peripheral_curves[peripheral_type][current_side][leave] -= 1
							next_tetrahedron, perm = current_tetrahedron.glued_to[leave]
							current_tetrahedron, current_side, arrive = next_tetrahedron, perm(current_side), perm(leave)
							break
			
			# Compute the algebraic intersection number between the longitude and meridian we just installed.
			# If the it is -1 then we need to reverse the direction of the meridian.
			# See SnapPy/kernel_code/intersection_numbers.c for how to do this.
			intersection_number = self.intersection_number(LONGITUDES, MERIDIANS, cusp)
			assert(abs(intersection_number) == 1)
			
			# We might need to reverse the orientation of one of these to get the right sign.
			# If the intersection number is -1 then we need to reverse the direction of one of them (we choose the meridian).
			if intersection_number < 0:
				for tetrahedron, side in cusp:
					for other in VERTICES_MEETING[side]:
						tetrahedron.peripheral_curves[MERIDIANS][side][other] = -tetrahedron.peripheral_curves[MERIDIANS][side][other]
	
	def slope(self):
		''' Return the slope of the peripheral curve in TEMPS relative to the set meridians and longitudes.
		
		Assumes that the meridian and longitude on this cusp have been set. '''
		
		longitude_intersection = self.intersection_number(LONGITUDES, TEMPS)
		meridian_intersection = self.intersection_number(MERIDIANS, TEMPS)
		
		# So the number of copies of the longitude and the meridian that we need to represent the path is:
		longitude_copies = -meridian_intersection
		meridian_copies = longitude_intersection
		
		return (meridian_copies, longitude_copies)
	
	def snappy_string(self, name='flipper triangulation', fillings=None):
		''' Return the SnapPy string describing this triangulation.
		
		If given, fillings is a list of slopes to perform Dehn filling along on each of the cusps.
		
		This triangulation must be closed. '''
		
		assert(isinstance(name, flipper.StringType))
		assert(fillings is None or isinstance(fillings, (list, tuple)))
		
		assert(self.is_closed())
		
		if fillings is None: fillings = [(0, 0)] * self.num_cusps
		
		strn = ''
		strn += '% Triangulation\n'
		strn += '%s\n' % name
		strn += 'not_attempted 0.0\n'
		strn += 'oriented_manifold\n'
		strn += 'CS_unknown\n'
		strn += '\n'
		strn += '%d 0\n' % self.num_cusps
		for slope in fillings:
			strn += '    torus %16.12f %16.12f\n' % slope
		
		strn += '\n'
		strn += '%d\n' % self.num_tetrahedra
		for tetrahedra in self:
			strn += tetrahedra.snappy_string() + '\n'
		
		return strn





''' The flipper GUI application. '''

from . import main
from . import pieces
from . import options
from . import inputbox
from . import choicebox
from . import progress
from . import widgets

# Set up shorter names for all of the different classes and some common constructors.
start = main.start
Options = options.Options
FlipperApplication = main.FlipperApplication
ColourPalette = pieces.ColourPalette
CanvasVertex = pieces.CanvasVertex
CanvasEdge = pieces.CanvasEdge
CanvasTriangle = pieces.CanvasTriangle
CurveComponent = pieces.CurveComponent
TrainTrackBlock = pieces.TrainTrackBlock
ProgressApp = progress.ProgressApp
SplitButton = widgets.SplitButton
Meter = widgets.Meter
AnimatedCanvas = widgets.AnimatedCanvas

lines_intersect = pieces.lines_intersect
interpolate = pieces.interpolate

apply_progression = progress.apply_progression
get_input = inputbox.get_input
get_choice = choicebox.get_choice





from __future__ import print_function

import flipper
import multiprocessing
from time import time

# We'll add a filter. Words missing b, c, or a and x are definitely not pseudo-Anosov.
# So lets filter those out as they are boring.
def filt(word):
	return \
		('a' in word or 'A' in word or 'x' in word or 'X' in word) and \
		('b' in word or 'B' in word) and \
		('c' in word or 'C' in word)

def main(length, cores):
	# Get an EquippedTriangulation.
	S = flipper.load('S_1_2')
	
	# Note this will work even if you want to set additional flags, such as conjugacy=False.
	
	print('Building words of length at most %d.' % length)
	# One can generate all words directly.
	start_time = time()
	all_words = list(S.all_words(length, filter=None))
	print('Built %d words.' % len(all_words))
	print('Completed in %0.3fs' % (time() - start_time))
	
	print('Starting again in parallel.')
	start_time = time()
	all_words2 = list(S.all_words(length, filter=None, cores=cores))
	print('Built %d words.' % len(all_words2))
	print('Completed in %0.3fs' % (time() - start_time))
	
	if len(all_words) == len(set(all_words)) and set(all_words) == set(all_words2):
		print('Lists match.')
	else:
		print('Lists DONT match.')

if __name__ == '__main__':
	import argparse
	parser = argparse.ArgumentParser(description='Generate all mapping classes of a given length')
	parser.add_argument('--length', type=int, default=7, help='length of words to generate')
	parser.add_argument('--cores', type=int, default=1, help='number of cores to use')
	parser.add_argument('--show', action='store_true', default=False, help='show the source code of this example and exit')
	args = parser.parse_args()
	
	if args.show:
		print(open(__file__, 'r').read())
	else:
		main(args.length, args.cores)

